# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example demonstrate how debug-log works"""

import asyncio

from juju.model import Model


async def main():
    model = Model()
    await model.connect_current()

    await model.debug_log(
        # limit=15,
        # exclude_module=['juju.worker.logforwarder'],
        # include_module=['juju.worker.dependency'], # <- only log dependency module
        # include=['machine-0'], # <- only log from machine-0
        # exclude=['machine-0'], # <- no log from machine-0
        # level='WARNING',
    )

    application = await model.deploy(
        "ch:ubuntu",
        application_name="ubuntu",
        series="trusty",
        channel="stable",
    )

    await model.wait_for_idle(status="active")

    await application.remove()
    await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

from juju.client.jujudata import FileJujuData
from juju.controller import Controller
from juju.errors import JujuError


class Juju:
    def __init__(self, jujudata=None):
        self.jujudata = jujudata or FileJujuData()

    def get_controllers(self):
        """Return list of all available controllers."""
        return self.jujudata.controllers()

    async def get_controller(self, name, include_passwords=False):
        """Get a controller by name.

        :param str name: Name of controller
        :param bool include_passwords: Include passwords for accounts

        The returned controller will try and connect to be ready to use.
        """
        # check if name is in the controllers.yaml
        controllers = self.jujudata.controllers()
        assert isinstance(controllers, dict)
        if name not in controllers:
            raise JujuError(
                "%s is not among the controllers: %s" % (name, controllers.keys())
            )

        # make a new Controller object that's connected to the
        # controller with the given name
        controller = Controller()
        await controller.connect(name)
        return controller



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Resets it
3. Deploys a charm and prints its config and constraints

"""

import asyncio
import logging

from juju.model import Model

log = logging.getLogger(__name__)

MB = 1


async def main():
    model = Model()
    # connect to current model with current user, per Juju CLI
    await model.connect()

    ubuntu_app = await model.deploy(
        "mysql",
        application_name="mysql",
        series="jammy",
        channel="edge",
        config={
            "cluster-name": "foo",
        },
        constraints={
            "mem": 256 * MB,
        },
    )
    await model.wait_for_idle(status="active")

    # update and check app config
    await ubuntu_app.set_config({"cluster-name": "bar"})
    config = await ubuntu_app.get_config()
    assert config["cluster-name"]["value"] == "bar"

    # update and check app constraints
    await ubuntu_app.set_constraints({"mem": 512 * MB})
    constraints = await ubuntu_app.get_constraints()
    assert constraints["mem"] == 512 * MB

    print("Removing mysql")
    await ubuntu_app.remove()

    print("Disconnecting from model")
    await model.disconnect()


if __name__ == "__main__":
    # logging.basicConfig(level=logging.DEBUG)
    ws_logger = logging.getLogger("websockets.protocol")
    ws_logger.setLevel(logging.INFO)
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
from __future__ import annotations

import copy
import logging
from typing import Any

import macaroonbakery.httpbakery as httpbakery
from packaging import version

from juju.client import client
from juju.client.connection import Connection
from juju.client.gocookies import GoCookieJar, go_to_py_cookie
from juju.client.jujudata import API_ENDPOINTS_KEY, FileJujuData
from juju.client.proxy.factory import proxy_from_config
from juju.errors import JujuConnectionError, JujuError, JujuUnknownVersion
from juju.version import CLIENT_VERSION

log = logging.getLogger("connector")


class NoConnectionException(Exception):
    """Raised by Connector when the connection method is called
    and there is no current connection.
    """

    pass


class Connector:
    """Abstracts out a reconnectable client that can connect
    to controllers and models found in the Juju data files.
    """

    def __init__(
        self,
        max_frame_size: int | None = None,
        bakery_client: Any | None = None,
        jujudata: Any | None = None,
    ):
        """Initialize a connector that will use the given parameters
        by default when making a new connection
        """
        self.max_frame_size = max_frame_size
        self.bakery_client = bakery_client
        self._connection = None
        self._log_connection = None
        self.controller_uuid = None
        self.model_name = None
        self.jujudata = jujudata or FileJujuData()

    def is_connected(self) -> bool:
        """Report whether there is a currently connected controller or not"""
        return self._connection is not None

    def connection(self) -> Connection:
        """Return the current connection; raises an exception if there
        is no current connection.
        """
        if not self.is_connected():
            raise NoConnectionException("not connected")
        assert self._connection
        return self._connection

    async def connect(self, **kwargs):
        """Connect to an arbitrary Juju model.

        kwargs are passed through to Connection.connect()
        """
        kwargs.setdefault("max_frame_size", self.max_frame_size)
        kwargs.setdefault("bakery_client", self.bakery_client)
        if "macaroons" in kwargs:
            if not kwargs["bakery_client"]:
                kwargs["bakery_client"] = httpbakery.Client()
            if not kwargs["bakery_client"].cookies:
                kwargs["bakery_client"].cookies = GoCookieJar()
            jar = kwargs["bakery_client"].cookies
            for macaroon in kwargs.pop("macaroons"):
                jar.set_cookie(go_to_py_cookie(macaroon))
        if "debug_log_conn" in kwargs:
            assert self._connection
            self._log_connection = await Connection.connect(**kwargs)
        else:
            # TODO (cderici): we need to investigate how to reuse/share
            # connections between Model & Controller objects
            # At least either avoid overwriting self._connection with
            # different types of connections (model, controller), or
            # have an indication of which type of entity this is
            # connected to.
            if self._connection:
                await self._connection.close()

            account = kwargs.pop("account", {})
            # Prioritize the username and password that user provided
            # If not enough, try to patch it with info from accounts.yaml
            if "username" not in kwargs and account.get("user"):
                kwargs.update(username=account.get("user"))
            if "password" not in kwargs and account.get("password"):
                kwargs.update(password=account.get("password"))

            if not ({"username", "password"}.issubset(kwargs)):
                required = {"username", "password"}.difference(kwargs)
                raise ValueError(
                    f"Some authentication parameters are required : {','.join(required)}"
                )
            self._connection = await Connection.connect(**kwargs)

        # Check if we support the target controller
        server_version = self._connection.info["server-version"]
        try:
            juju_server_version = version.parse(server_version)
        except version.InvalidVersion as err:
            # We're only interested in the major version, so
            # we attempt to clean up versions such as 3.4-rc1.2 as just 3.4
            if "-" not in server_version:
                raise JujuUnknownVersion(err)
            juju_server_version = version.parse(server_version.split("-")[0])

        # CLIENT_VERSION statically comes from the VERSION file in the repo
        client_version = version.parse(CLIENT_VERSION)

        if juju_server_version.major != client_version.major:
            raise JujuConnectionError(
                "juju server-version %s not supported" % juju_server_version
            )

    async def disconnect(self, entity):
        """Shut down the watcher task and close websockets."""
        if self._connection:
            log.debug(f"Connector: closing {entity} connection")
            await self._connection.close()
            self._connection = None
        if self._log_connection:
            log.debug("Also closing debug-log connection")
            await self._log_connection.close()
            self._log_connection = None

    async def connect_controller(
        self, controller_name=None, specified_facades=None, **kwargs
    ):
        """Connect to a controller by name. If the name is empty, it
        connect to the current controller.
        """
        if not controller_name:
            controller_name = self.jujudata.current_controller()

        controller = self.jujudata.controllers()[controller_name]
        endpoints = controller[API_ENDPOINTS_KEY]
        accounts = self.jujudata.accounts().get(controller_name, {})

        proxy = proxy_from_config(controller.get("proxy-config", None))

        kwargs.update(
            endpoint=endpoints,
            uuid=None,
            account=accounts,
            cacert=controller.get("ca-cert"),
            bakery_client=self.bakery_client_for_controller(controller_name),
            specified_facades=specified_facades,
            proxy=proxy,
        )
        await self.connect(**kwargs)
        self.controller_name = controller_name
        self.controller_uuid = controller["uuid"]

    async def connect_model(self, _model_name=None, **kwargs):
        """Connect to a model by name. If either controller or model
        parts of the name are empty, the current controller and/or model
        will be used.

        :param str model: <controller>:<model>
        """
        try:
            controller_name, _model_name = self.jujudata.parse_model(_model_name)
            controller = self.jujudata.controllers().get(controller_name)
        except JujuError as e:
            raise JujuConnectionError(e.message) from e
        if controller is None:
            raise JujuConnectionError(f"Controller {controller_name} not found")
        endpoints = controller[API_ENDPOINTS_KEY]
        account = self.jujudata.accounts().get(controller_name, {})
        models = self.jujudata.models().get(controller_name, {}).get("models", {})
        model_uuid = None
        if _model_name in models:
            model_uuid = models[_model_name]["uuid"]
        else:
            # let's try to find it through the actual controller
            await self.connect_controller(controller_name=controller_name)
            # get the facade
            controller_facade = client.ControllerFacade.from_connection(
                self.connection()
            )
            # get all the user models from the api
            response = await controller_facade.AllModels()
            # search the one that contains admin/model_name
            for user_model in response.user_models:
                if "admin/" + user_model.model.name == _model_name:
                    model_uuid = user_model.model.uuid

        if model_uuid is None:
            raise JujuConnectionError(f"Model not found: {_model_name}")

        proxy = proxy_from_config(controller.get("proxy-config", None))

        # TODO remove the need for base.CleanModel to subclass
        # JujuData.
        kwargs.update(
            endpoint=endpoints,
            uuid=model_uuid,
            account=account,
            cacert=controller.get("ca-cert"),
            bakery_client=self.bakery_client_for_controller(controller_name),
            proxy=proxy,
        )
        await self.connect(**kwargs)
        # TODO this might be a good spot to trigger refreshing the
        # local cache (the connection to the model might help)
        self.model_name = controller_name + ":" + _model_name
        return model_uuid

    def bakery_client_for_controller(self, controller_name):
        """Make a copy of the bakery client with a the appropriate controller's
        cookiejar in it.
        """
        bakery_client = self.bakery_client
        if bakery_client:
            bakery_client = copy.copy(bakery_client)
        else:
            bakery_client = httpbakery.Client()
        bakery_client.cookies = self.jujudata.cookies_for_controller(controller_name)
        return bakery_client



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current controller
2. Creates two models for consuming and offering
2. Deploys a charm and waits until it reports itself active
3. Creates an offer
4. Lists the offer
5. Deploys a charm and adds relation to the offering url
6. Destroys the units and applications

"""

import asyncio
import tempfile
import time
from logging import getLogger

from juju.controller import Controller

log = getLogger(__name__)


async def main():
    controller = Controller()
    print("Connecting to controller")
    await controller.connect()

    try:
        print("Creating models")
        offering_model = await controller.add_model("test-cmr-1")
        consuming_model = await controller.add_model("test-cmr-2")

        print("Deploying mysql")
        await offering_model.deploy(
            "ch:mysql",
            application_name="mysql",
            series="jammy",
            channel="edge",
        )

        print("Waiting for active")
        await offering_model.wait_for_idle(status="active")

        print("Adding offer")
        await offering_model.create_offer("mysql:db")

        offers = await offering_model.list_offers()
        await offering_model.block_until(
            lambda: all(offer.application_name == "mysql" for offer in offers.results)
        )

        print(
            "Show offers",
            ", ".join(
                "%s: %s" % item
                for offer in offers.results
                for item in vars(offer).items()
            ),
        )

        # TODO (cderici): wordpress charm is somewhat problematic in 3.0,
        #  this example needs to be revisited.
        print("Deploying wordpress")
        application_2 = await consuming_model.deploy(
            "ch:trusty/wordpress",
            application_name="wordpress",
            series="xenial",
            channel="stable",
        )

        print("Waiting for executing")
        await consuming_model.block_until(
            lambda: all(
                unit.agent_status == "executing" for unit in application_2.units
            )
        )

        await consuming_model.relate("wordpress", "admin/test-cmr-1.mysql")

        print("Exporting bundle")
        with tempfile.TemporaryDirectory() as dirpath:
            await offering_model.export_bundle(f"{dirpath}/bundle.yaml")

        time.sleep(10)

        print("Remove SAAS")
        await consuming_model.remove_saas("mysql")

        print("Removing offer")
        await offering_model.remove_offer("admin/test-cmr-1.mysql", force=True)

        print("Destroying models")
        await controller.destroy_model(offering_model.info.uuid)
        await controller.destroy_model(consuming_model.info.uuid)

    except Exception:
        log.exception("Example failed!")
        raise

    finally:
        print("Disconnecting from controller")
        await controller.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

from juju.client.proxy.kubernetes.proxy import KubernetesProxy


def proxy_from_config(conf):
    if conf is None:
        return None

    if "type" not in conf:
        return None

    proxy_type = conf["type"]
    if proxy_type != "kubernetes-port-forward":
        raise ValueError("unknown proxy type %s" % proxy_type)

    return _construct_kube_proxy(conf["config"])


def _construct_kube_proxy(config):
    return KubernetesProxy(
        config.get("api-host", ""),
        config.get("namespace", ""),
        config.get("remote-port", ""),
        config.get("service", ""),
        config.get("service-account-token", ""),
        config.get("ca-cert", None),
    )



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

#
# Module that parses constraints
#
# The current version of juju core expects the client to take
# constraints given in the form "mem=10G foo=bar" and parse them into
# json that looks like {"mem": 10240, "foo": "bar"}. This module helps us
# accomplish that task.
#
# We do not attempt to duplicate the checking done in
# client/_client.py:Value here. That class will verify that the
# constraints keys are valid, and that we can successfully dump the
# constraints dict to json.
#
# Once https://bugs.launchpad.net/juju/+bug/1645402 is addressed, this
# module should be deprecated.
#

import re
from typing import Dict, List, Mapping, Optional, TypedDict, Union

from typing_extensions import NotRequired, Required

# Matches on a string specifying memory size
MEM = re.compile("^[1-9][0-9]*[MGTP]$")

# Multiplication factors to get Megabytes
# https://github.com/juju/juju/blob/master/constraints/constraints.go#L666
FACTORS = {
    "M": 1024**0,
    "G": 1024**1,
    "T": 1024**2,
    "P": 1024**3,
    "E": 1024**4,
    "Z": 1024**5,
    "Y": 1024**6,
}

# List of supported constraint keys, see
# http://github.com/cderici/juju/blob/2.9/core/constraints/constraints.go#L20-L39
SUPPORTED_KEYS = [
    "arch",
    "container",
    "cpu_cores",
    "cores",
    "cpu_power",
    "mem",
    "root_disk",
    "root_disk_source",
    "tags",
    "instance_role",
    "instance_type",
    "spaces",
    "virt_type",
    "zones",
    "allocate_public_ip",
]

LIST_KEYS = {"tags", "spaces", "zones"}

SNAKE1 = re.compile(r"(.)([A-Z][a-z]+)")
SNAKE2 = re.compile("([a-z0-9])([A-Z])")


ParsedValue = Union[int, bool, str]


class ConstraintsDict(TypedDict, total=False):
    allocate_public_ip: ParsedValue
    arch: ParsedValue
    container: ParsedValue
    cores: ParsedValue
    cpu_cores: ParsedValue
    cpu_power: ParsedValue
    instance_role: ParsedValue
    instance_type: ParsedValue
    mem: ParsedValue
    root_disk: ParsedValue
    root_dist_source: ParsedValue
    spaces: List[ParsedValue]
    tags: List[ParsedValue]
    virt_type: ParsedValue
    zones: List[ParsedValue]


def parse(constraints: Union[str, ConstraintsDict]) -> Optional[ConstraintsDict]:
    """Constraints must be expressed as a string containing only spaces
    and key value pairs joined by an '='.

    """
    if not constraints:
        return None

    if isinstance(constraints, dict):
        # Forwards compatibility: already parsed
        return constraints

    normalized_constraints: ConstraintsDict = {}
    for s in constraints.split(" "):
        if "=" not in s:
            raise ValueError("malformed constraint %s" % s)

        k, v = s.split("=")
        normalized_constraints[normalize_key(k)] = (
            normalize_list_value(v) if k in LIST_KEYS else normalize_value(v)
        )

    return normalized_constraints


def normalize_key(orig_key: str) -> str:
    key = orig_key.strip()

    key = key.replace("-", "_")  # Our _client lib wants "_" in place of "-"

    # Convert camelCase to snake_case
    key = SNAKE1.sub(r"\1_\2", key)
    key = SNAKE2.sub(r"\1_\2", key).lower()

    if key not in SUPPORTED_KEYS:
        raise ValueError("unknown constraint in %s" % orig_key)
    return key


def normalize_value(value: str) -> Union[int, bool, str]:
    value = value.strip()

    if MEM.match(value):
        # Translate aliases to Megabytes. e.g. 1G = 10240
        return int(value[:-1]) * FACTORS[value[-1:]]

    if value.isdigit():
        return int(value)

    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False

    return value


def normalize_list_value(value: str) -> List[ParsedValue]:
    values = value.strip().split(",")
    return [normalize_value(value) for value in values]


STORAGE = re.compile(
    # original regex:
    # '(?:(?:^|(?<=,))(?:|(?P<pool>[a-zA-Z]+[-?a-zA-Z0-9]*)|(?P<count>-?[0-9]+)|(?:(?P<size>-?[0-9]+(?:\\.[0-9]+)?)(?P<size_exp>[MGTPEZY])(?:i?B)?))(?:$|,))'
    # with formatting and explanation -- note that this regex is used with re.finditer:
    "(?:"
    "(?:^|(?<=,))"  # start of string or previous match ends with ','
    "(?:"  # match one of the following:
    "|(?P<pool>[a-zA-Z]+[-?a-zA-Z0-9]*)"  # * pool: a sequence starting with a letter, ending with a letter or number,
    # ------- and including letters, numbers and hyphens (no more than one in a row)
    "|(?P<count>-?[0-9]+)"  # * count: an optional minus sign followed by one or more digits
    "|(?:"  # * size (number) and size_exp (units):
    "(?P<size>-?[0-9]+(?:\\.[0-9]+)?)"  # -- * an optional minus sign followed by one or more digits, optionally with decimal point and more digits
    "(?P<size_exp>[MGTPEZY])(?:i?B)?)"  # -- * one of MGTPEZY, optionally followed by iB or B, for example 1M or 2.0MB or -3.3MiB
    ")"
    "(?:$|,)"  # end of string or ','
    ")"
)


class StorageConstraintDict(TypedDict):
    count: Required[int]  # >= 1
    pool: NotRequired[str]
    size: NotRequired[int]


def parse_storage_constraint(constraint: str) -> StorageConstraintDict:
    storage: StorageConstraintDict = {"count": 1}
    for m in STORAGE.finditer(constraint):
        pool = m.group("pool")
        if pool:
            if "pool" in storage:
                raise ValueError("pool already specified")
            storage["pool"] = pool
        count = m.group("count")
        if count:
            count = int(count)
            storage["count"] = count if count > 0 else 1
        size = m.group("size")
        if size:
            storage["size"] = int(float(size) * FACTORS[m.group("size_exp")])
    return storage


def parse_storage_constraints(
    constraints: Optional[Mapping[str, Union[str, StorageConstraintDict]]] = None,
) -> Dict[str, StorageConstraintDict]:
    if constraints is None:
        return {}
    parsed: dict[str, StorageConstraintDict] = {}
    for label, storage_constraint in constraints.items():
        if isinstance(storage_constraint, str):
            parsed[label] = parse_storage_constraint(storage_constraint)
        elif isinstance(storage_constraint, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
            parsed[label] = storage_constraint
        else:
            raise ValueError(
                f"Unexpected constraint {storage_constraint!r}"
                f" for label {label!r} in {constraints}"
            )
    return parsed


DEVICE = re.compile(
    # original regex:
    # '^(?P<count>[0-9]+)?(?:^|,)(?P<type>[^,]+)(?:$|,(?!$))(?P<attrs>(?:[^=]+=[^;]+)+)*$'
    # with formatting and explanation -- note this regex is used with re.match:
    "^"  # start of string
    "(?P<count>[0-9]+)?"  # count is 1+ digits, and is optional
    "(?:^|,)"  # match start of string or a comma
    # -- so type can be at the start or comma separated from count
    "(?P<type>[^,]+)"  # type is 1+ anything not a comma (including digits), and is required
    "(?:$|,(?!$))"  # match end of string | or a non-trailing comma
    # -- so type can be at the end or followed by attrs
    "(?P<attrs>(?:[^=]+=[^;]+)+)*"  # attrs is any number of semicolon separated key=value items
    # -- value can have spare '=' inside, possible not intended
    # -- attrs will be matched with ATTR.finditer afterwards in parse_device_constraint
    "$"  # end of string
)
ATTR = re.compile(";?(?P<key>[^=]+)=(?P<value>[^;]+)")


class DeviceConstraintDict(TypedDict):
    count: Required[int]
    type: Required[str]
    attributes: NotRequired[Dict[str, str]]


def parse_device_constraint(constraint: str) -> DeviceConstraintDict:
    m = DEVICE.match(constraint)
    if m is None:
        raise ValueError("device constraint does not match")
    device: DeviceConstraintDict = {}
    count = m.group("count")
    if count:
        count = int(count)
        device["count"] = count if count > 0 else 1
    else:
        device["count"] = 1
    device["type"] = m.group("type")
    attrs = m.group("attrs")
    if attrs:
        device["attributes"] = {
            match.group("key"): match.group("value") for match in ATTR.finditer(attrs)
        }
    return device



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.


"""This example:

1. Connects to the current model
2. Creates a machine
3. Waits for the machine agent to start and prints out the reported machine
   hostname.

NOTE: this example requires a 2.8.10+ controller.
"""

import asyncio
import logging

from juju.model import Model

MB = 1
GB = 1024


async def main():
    model = Model()
    await model.connect()

    try:
        # Add a machine and wait until the machine agents starts
        machine1 = await model.add_machine()
        await model.block_until(lambda: machine1.agent_status == "started")

        # At this point we can access the reported hostname via the hostname
        # property of the machine model.
        print(f"machine1 hostname: {machine1.hostname}")

        await machine1.destroy(force=True)
    finally:
        await model.disconnect()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    ws_logger = logging.getLogger("websockets.protocol")
    ws_logger.setLevel(logging.INFO)

    asyncio.run(main())



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client import (
    _client1,
    _client2,
    _client3,
    _client4,
    _client5,
    _client6,
    _client7,
    _client8,
    _client9,
    _client10,
    _client11,
    _client12,
    _client17,
    _client19,
    _client20,
)
from juju.client._definitions import *

CLIENTS = {
    "7": _client7,
    "3": _client3,
    "4": _client4,
    "2": _client2,
    "17": _client17,
    "6": _client6,
    "11": _client11,
    "1": _client1,
    "10": _client10,
    "9": _client9,
    "5": _client5,
    "19": _client19,
    "20": _client20,
    "8": _client8,
    "12": _client12,
}


def lookup_facade(name, version):
    """Given a facade name and version, attempt to pull that facade out
    of the correct client<version>.py file.

    """
    for _version in range(int(version), 0, -1):
        try:
            facade = getattr(CLIENTS[str(_version)], name)
            return facade
        except (KeyError, AttributeError):
            continue
    else:
        raise ImportError(f"No supported version for facade: {name}")


class TypeFactory:
    @classmethod
    def from_connection(cls, connection):
        """Given a connected Connection object, return an initialized and
        connected instance of an API Interface matching the name of
        this class.

        @param connection: initialized Connection object.

        """
        facade_name = cls.__name__
        if not facade_name.endswith("Facade"):
            raise TypeError(f"Unexpected class name: {facade_name}")
        facade_name = facade_name[: -len("Facade")]
        version = connection.facades.get(facade_name)
        if version is None:
            raise Exception(f"No facade {facade_name} in facades {connection.facades}")

        c = lookup_facade(cls.__name__, version)
        c = c()
        c.connect(connection)

        return c

    @classmethod
    def best_facade_version(cls, connection):
        """Returns the best facade version for a given facade. This will help with
        trying to provide different functionality for different facade versions.

        @param connection: initialized Connection object.
        """
        facade_name = cls.__name__
        if not facade_name.endswith("Facade"):
            raise TypeError(f"Unexpected class name: {facade_name}")
        facade_name = facade_name[: -len("Facade")]
        return connection.facades.get(facade_name)


class ActionFacade(TypeFactory):
    pass


class AdminFacade(TypeFactory):
    pass


class AllModelWatcherFacade(TypeFactory):
    pass


class AllWatcherFacade(TypeFactory):
    pass


class AnnotationsFacade(TypeFactory):
    pass


class ApplicationFacade(TypeFactory):
    pass


class ApplicationOffersFacade(TypeFactory):
    pass


class BackupsFacade(TypeFactory):
    pass


class BlockFacade(TypeFactory):
    pass


class BundleFacade(TypeFactory):
    pass


class CharmsFacade(TypeFactory):
    pass


class ClientFacade(TypeFactory):
    pass


class CloudFacade(TypeFactory):
    pass


class ControllerFacade(TypeFactory):
    pass


class CredentialManagerFacade(TypeFactory):
    pass


class FacadeVersions(TypeFactory):
    pass


class FirewallRulesFacade(TypeFactory):
    pass


class HighAvailabilityFacade(TypeFactory):
    pass


class ImageMetadataManagerFacade(TypeFactory):
    pass


class KeyManagerFacade(TypeFactory):
    pass


class MachineManagerFacade(TypeFactory):
    pass


class MetricsDebugFacade(TypeFactory):
    pass


class ModelConfigFacade(TypeFactory):
    pass


class ModelGenerationFacade(TypeFactory):
    pass


class ModelManagerFacade(TypeFactory):
    pass


class ModelUpgraderFacade(TypeFactory):
    pass


class PayloadsFacade(TypeFactory):
    pass


class PingerFacade(TypeFactory):
    pass


class ResourcesFacade(TypeFactory):
    pass


class SSHClientFacade(TypeFactory):
    pass


class SecretBackendsFacade(TypeFactory):
    pass


class SecretsFacade(TypeFactory):
    pass


class SpacesFacade(TypeFactory):
    pass


class StorageFacade(TypeFactory):
    pass


class SubnetsFacade(TypeFactory):
    pass


class UserManagerFacade(TypeFactory):
    pass



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio

import hvac

from juju.model import Model


async def main():
    """This is a complete example that deploys vault, uses a
    vault client to initialize it, and registers the backend.
    """
    m = Model()
    await m.connect()

    # # deploy postgresql
    await m.deploy("postgresql", series="focal")
    # # deploy vault
    await m.deploy("vault", series="focal")
    # # relate/integrate
    await m.integrate("vault:db", "postgresql:db")
    # # wait for the
    await m.wait_for_idle(["postgresql", "vault"])
    # # expose vault
    vault_app = m.applications["vault"]
    await vault_app.expose()

    # Get a vault client
    # Deploy this entire thing
    status = await m.get_status()
    target = ""
    for unit in status.applications["vault"].units.values():
        target = unit.public_address

    vault_url = "http://%s:8200" % target
    vault_client = hvac.Client(url=vault_url)

    # Initialize vault
    keys = vault_client.sys.initialize(3, 2)
    print(keys)

    # Unseal vault
    vault_client.sys.submit_unseal_keys(keys["keys"])

    target_unit = m.applications["vault"].units[0]
    action = await target_unit.run_action("authorize-charm", token=keys["root_token"])
    await action.wait()

    # Add the secret backend
    c = await m.get_controller()
    response = await c.add_secret_backends(
        "1111",
        "examplevault",
        "vault",
        {"endpoint": vault_url, "token": keys["root_token"]},
    )
    print("Output from add secret backends")
    print(response["results"])

    # List the secrets backend
    result = await c.list_secret_backends()
    print("Output from list secret backends")
    print(result["results"])

    # Remove it
    await c.remove_secret_backends("examplevault")

    # # Finally after removing
    result = await c.list_secret_backends()
    print("Output from list secret backends after removal")
    print(result["results"])

    await m.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Starts an AllWatcher
3. Prints all changes received from the AllWatcher
4. Runs forever (kill with Ctrl-C)

"""

import asyncio
import logging

from juju.controller import Controller


async def watch():
    controller = Controller()
    # connect to current
    # controller with current user, per Juju CLI
    await controller.connect()

    # Need to call the WatchModelSummaries or WatchAllModelSummaries on the
    # controller.
    def callback(summary):
        print(f"-- change --\n{summary}\n")

    await controller.watch_model_summaries(callback)

    while True:
        await asyncio.sleep(1)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    ws_logger = logging.getLogger("websockets.protocol")
    ws_logger.setLevel(logging.INFO)
    logging.getLogger("juju.client.connection").setLevel(logging.INFO)
    # Run loop until the process is manually stopped (watch will loop
    # forever).
    asyncio.run(watch())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This is a very basic example that connects to the currently selected model
and prints the number of applications deployed to it.
"""

import asyncio
import logging

from juju.model import Model

log = logging.getLogger(__name__)


async def main():
    model = Model()
    try:
        # connect to the current model with the current user, per the Juju CLI
        await model.connect()
        print(f"There are {len(model.applications)} applications")
    finally:
        if model.is_connected():
            print("Disconnecting from model")
            await model.disconnect()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to current controller.
2. Gets the cloud name in which the controller lives on
3. Disconnects from the controller

"""

import asyncio
import logging

from juju.controller import Controller


async def main():
    controller = Controller()
    # connect to current controller with current user, per Juju CLI
    await controller.connect()

    await controller.get_cloud()

    await controller.disconnect()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    ws_logger = logging.getLogger("websockets.protocol")
    ws_logger.setLevel(logging.INFO)
    asyncio.run(main())



#!/usr/bin/env python3

# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example shows how to reconnect to a model if you encounter an error

1. Connects to current model.
2. Attempts to get an application that doesn't exist.
3. Disconnect then reconnect.

"""

import asyncio

from juju.errors import JujuEntityNotFoundError
from juju.model import Model


async def main():
    model = Model()

    retries = 3
    for _ in range(0, retries):
        await model.connect_current()
        try:
            print(model.applications["foo"].relations)
        except JujuEntityNotFoundError as e:
            print(e.entity_name)
        finally:
            await model.disconnect()
        # Everything worked out, continue on wards.


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Deploy a charm and waits until it reports itself active
3. Destroys the unit and application

"""

import asyncio

from juju.model import Model


async def main():
    model = Model()
    print("Connecting to model")
    await model.connect()

    try:
        print("Deploying ")
        application = await model.deploy(
            "ch:juju-qa-test",
            application_name="juju-qa-test",
            channel="2.0/edge",
        )

        print("Waiting for active")
        await model.block_until(
            lambda: all(unit.workload_status == "active" for unit in application.units)
        )

    finally:
        print("Disconnecting from model")
        await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
import logging
import os
import tempfile

from kubernetes import client
from kubernetes.stream import portforward

from juju.client.proxy.proxy import Proxy, ProxyNotConnectedError

log = logging.getLogger("juju.client.connection")


class KubernetesProxy(Proxy):
    def __init__(
        self,
        api_host,
        namespace,
        remote_port,
        service,
        service_account_token,
        ca_cert=None,
    ):
        config = client.Configuration()
        config.host = api_host
        config.ssl_ca_cert = ca_cert
        config.api_key = {"authorization": "Bearer " + service_account_token}

        self.namespace = namespace
        self.remote_port = remote_port
        self.service = service
        self.temp_ca_path = None
        self.port_forwarder = None

        try:
            self.remote_port = int(remote_port)
        except ValueError:
            raise ValueError(f"Invalid port number: {remote_port}")

        if ca_cert:
            with tempfile.NamedTemporaryFile(delete=False) as f:
                f.write(bytes(ca_cert, "utf-8"))
            self.temp_ca_path = f.name
            config.ssl_ca_cert = f.name

        self.api_client = client.ApiClient(config)

    def connect(self):
        corev1 = client.CoreV1Api(self.api_client)
        service = corev1.read_namespaced_service(self.service, self.namespace)

        label_selector = ",".join(k + "=" + v for k, v in service.spec.selector.items())

        pods = corev1.list_namespaced_pod(
            namespace=self.namespace,
            label_selector=label_selector,
        )

        self.port_forwarder = portforward(
            corev1.connect_get_namespaced_pod_portforward,
            pods.items[0].metadata.name,
            self.namespace,
            ports=str(self.remote_port),
        )

    def __del__(self):
        self.close()
        if self.temp_ca_path:
            os.unlink(self.temp_ca_path)

    def close(self):
        if self.port_forwarder:
            self.port_forwarder.close()
            self.port_forwarder = None

    def socket(self):
        if self.port_forwarder is not None:
            return self.port_forwarder.socket(self.remote_port)._socket
        raise ProxyNotConnectedError()



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import datetime
import http.cookiejar as cookiejar
import json
import time

from backports.datetime_fromisoformat import datetime_fromisoformat


class GoCookieJar(cookiejar.FileCookieJar):
    """A CookieJar implementation that reads and writes cookies
    to the cookiejar format as understood by the Go package
    github.com/juju/persistent-cookiejar.
    """

    def _really_load(self, f, filename, ignore_discard, ignore_expires):
        """Implement the _really_load method called by FileCookieJar
        to implement the actual cookie loading
        """
        data = json.load(f) or []
        now = time.time()
        for cookie in map(go_to_py_cookie, data):
            if not ignore_expires and cookie.is_expired(now):
                continue
            self.set_cookie(cookie)

    def save(self, filename=None, ignore_discard=False, ignore_expires=False):
        """Implement the FileCookieJar abstract method."""
        if filename is None:
            if self.filename is not None:
                filename = self.filename
            else:
                raise ValueError(cookiejar.MISSING_FILENAME_TEXT)

        # TODO: obtain file lock, read contents of file, and merge with
        # current content.
        go_cookies = []
        now = time.time()
        for cookie in self:
            if not ignore_discard and cookie.discard:
                continue
            if not ignore_expires and cookie.is_expired(now):
                continue
            go_cookies.append(py_to_go_cookie(cookie))
        with open(filename, "w") as f:
            f.write(json.dumps(go_cookies))


def go_to_py_cookie(go_cookie):
    """Convert a Go-style JSON-unmarshaled cookie into a Python cookie"""
    expires = None
    if go_cookie.get("Expires") is not None:
        t = datetime_fromisoformat(go_cookie["Expires"])
        expires = t.timestamp()
    return cookiejar.Cookie(
        version=0,
        name=go_cookie["Name"],
        value=go_cookie["Value"],
        port=None,
        port_specified=False,
        # Unfortunately Python cookies don't record the original
        # host that the cookie came from, so we'll just use Domain
        # for that purpose, and record that the domain was specified,
        # even though it probably was not. This means that
        # we won't correctly record the CanonicalHost entry
        # when writing the cookie file after reading it.
        domain=go_cookie["Domain"],
        domain_specified=not go_cookie["HostOnly"],
        domain_initial_dot=False,
        path=go_cookie["Path"],
        path_specified=True,
        secure=go_cookie["Secure"],
        expires=expires,
        discard=False,
        comment=None,
        comment_url=None,
        rest=None,
        rfc2109=False,
    )


def py_to_go_cookie(py_cookie):
    """Convert a python cookie to the JSON-marshalable Go-style cookie form."""
    # TODO (perhaps):
    #   HttpOnly
    #   Creation
    #   LastAccess
    #   Updated
    # not done properly: CanonicalHost.
    go_cookie = {
        "Name": py_cookie.name,
        "Value": py_cookie.value,
        "Domain": py_cookie.domain,
        "HostOnly": not py_cookie.domain_specified,
        "Persistent": not py_cookie.discard,
        "Secure": py_cookie.secure,
        "CanonicalHost": py_cookie.domain,
    }
    if py_cookie.path_specified:
        go_cookie["Path"] = py_cookie.path
    if py_cookie.expires is not None:
        # Note: fromtimestamp bizarrely produces a time without
        # a time zone, so we need to use accept_naive.
        go_cookie["Expires"] = datetime.datetime.fromtimestamp(
            py_cookie.expires
        ).isoformat()
    return go_cookie



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

from enum import Enum

from . import utils
from .errors import JujuError


class Source(Enum):
    """Source defines a origin source. Providing a hint to the controller about
    what the charm identity is from the URL and origin source.

    """

    LOCAL = "local"
    CHARM_HUB = "charm-hub"

    def __str__(self):
        return self.value


class Origin:
    def __init__(self, source, channel, platform):
        self.source = source
        self.channel = channel
        self.platform = platform

    def __str__(self):
        return f"origin using source {self.source!s} for channel {self.channel} and platform {self.platform}"


class Risk(Enum):
    STABLE = "stable"
    CANDIDATE = "candidate"
    BETA = "beta"
    EDGE = "edge"

    def __str__(self):
        return self.value

    @staticmethod
    def valid(potential):
        for risk in [Risk.STABLE, Risk.CANDIDATE, Risk.BETA, Risk.EDGE]:
            if str(risk) == potential:
                return True
        return False


class Channel:
    """Channel identifies and describes completely a store channel.

    A channel consists of, and is subdivided by, tracks, risk-levels and
     - Tracks enable snap developers to publish multiple supported releases of
       their application under the same snap name.
     - Risk-levels represent a progressive potential trade-off between stability
       and new features.

    The complete channel name can be structured as three distinct parts separated
    by slashes:

       <track>/<risk>

    """

    def __init__(self, track=None, risk=None):
        if not Risk.valid(risk):
            raise JujuError(f"unexpected risk {risk}")

        self.track = track or ""
        self.risk = risk

    @staticmethod
    def parse(s: str):
        """Parse a channel from a given string.
        Parse does not take into account branches.

        """
        if not s:
            raise JujuError("channel cannot be empty")

        p = s.split("/")

        risk = None
        track = None
        if len(p) == 1:
            if Risk.valid(p[0]):
                risk = p[0]
            else:
                track = p[0]
                risk = str(Risk.STABLE)
        elif len(p) == 2:
            track = p[0]
            risk = p[1]
        else:
            raise JujuError(f"channel is malformed and has too many components {s}")

        if risk is not None and not Risk.valid(risk):
            raise JujuError(f"risk in channel {s} is not valid")
        if track is not None and track == "":
            raise JujuError(f"track in channel {s} is not valid")

        return Channel(track, risk)

    def normalize(self):
        track = self.track if self.track != "" else ""
        risk = self.risk if self.risk != "" else ""
        return Channel(track, risk)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.track == other.track and self.risk == other.risk
        return False

    def __str__(self):
        path = self.risk
        if self.track != "":
            path = f"{self.track}/{path}"
        return path

    def compute_base_channel(self, series):
        """Determines the channel for a client.Base
        A base channel is a track/risk/branch

        """
        _ch = [self.risk]
        tr = utils.get_series_version(series)
        if tr:
            _ch = [tr, *_ch]
        return "/".join(_ch)


class Platform:
    """ParsePlatform parses a string representing a store platform.
    Serialized version of platform can be expected to conform to the following:

     1. Architecture is mandatory.
     2. OS is optional and can be dropped. Series is mandatory if OS wants
     to be displayed.
     3. Series is also optional.

    To indicate something is missing `unknown` can be used in place.

    Examples:
     1. `<arch>/<os>/<series>`
     2. `<arch>`
     3. `<arch>/<series>`
     4. `<arch>/unknown/<series>`

    """

    def __init__(self, arch, series=None, os=None):
        self.arch = arch
        self.series = series
        self.os = os

    @staticmethod
    def parse(s):
        if not s:
            raise JujuError("platform cannot be empty")

        p = s.split("/")

        arch = None
        os = None
        series = None
        if len(p) == 1:
            arch = p[0]
        elif len(p) == 2:
            arch = p[0]
            series = p[1]
        elif len(p) == 3:
            arch = p[0]
            os = p[1]
            series = p[2]
        else:
            raise JujuError(f"platform is malformed and has too many components {s}")

        if not arch:
            raise JujuError(f"architecture in platform {s} is not valid")
        if os is not None and os == "":
            raise JujuError(f"os in platform {s} is not valid")
        if series is not None and series == "":
            raise JujuError(f"series in platform {s} is not valid")

        return Platform(arch, series, os)

    def normalize(self):
        os = self.os if self.os is not None or self.os != "unknown" else None
        series = self.series
        if series is None or series == "unknown":
            os = None
            series = None

        return Platform(self.arch, series, os)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.arch == other.arch
                and self.os == other.os
                and self.series == other.series
            )
        return False

    def __str__(self):
        path = self.arch
        if self.os is not None and self.os != "":
            path = f"{path}/{self.os}"
        if self.series is not None and self.series != "":
            path = f"{path}/{self.series}"
        return path



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class ClientFacade(Type):
    name = "Client"
    version = 8

    @ReturnMapping(FullStatus)
    async def FullStatus(self, include_storage=None, patterns=None):
        """FullStatus gives the information needed for juju status over the api

        include_storage : bool
        patterns : typing.Sequence[str]
        Returns -> FullStatus
        """
        if include_storage is not None and not isinstance(include_storage, bool):
            raise Exception(
                f"Expected include_storage to be a bool, received: {type(include_storage)}"
            )

        if patterns is not None and not isinstance(patterns, (bytes, str, list)):
            raise Exception(
                f"Expected patterns to be a Sequence, received: {type(patterns)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="FullStatus", version=8, params=_params)
        _params["include-storage"] = include_storage
        _params["patterns"] = patterns
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StatusHistoryResults)
    async def StatusHistory(self, requests=None):
        """StatusHistory returns a slice of past statuses for several entities.

        requests : typing.Sequence[~StatusHistoryRequest]
        Returns -> StatusHistoryResults
        """
        if requests is not None and not isinstance(requests, (bytes, str, list)):
            raise Exception(
                f"Expected requests to be a Sequence, received: {type(requests)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="StatusHistory", version=8, params=_params)
        _params["requests"] = requests
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(AllWatcherId)
    async def WatchAll(self):
        """WatchAll initiates a watcher for entities in the connected model.

        Returns -> AllWatcherId
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="WatchAll", version=8, params=_params)

        reply = await self.rpc(msg)
        return reply



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
from __future__ import annotations

# TODO: Tags should be a proper class, so that we can distinguish whether
# something is already a tag or not.  For example, 'user-foo' is a valid
# username, but is ambiguous with the already-tagged username 'foo'.


def _prefix(prefix: str, s: str) -> str:
    if s and not s.startswith(prefix):
        return f"{prefix}{s}"
    return s


def untag(prefix: str, s: str) -> str:
    if s and s.startswith(prefix):
        return s[len(prefix) :]
    return s


def cloud(cloud_name: str) -> str:
    return _prefix("cloud-", cloud_name)


def controller(controller_uuid: str) -> str:
    return _prefix("controller-", controller_uuid)


def credential(cloud: str, user: str, credential_name: str) -> str:
    credential_string = f"{cloud}_{user}_{credential_name}"
    return _prefix("cloudcred-", credential_string)


def model(model_uuid: str) -> str:
    return _prefix("model-", model_uuid)


def machine(machine_id: str) -> str:
    return _prefix("machine-", machine_id)


def user(username: str) -> str:
    return _prefix("user-", username)


def application(app_name: str) -> str:
    return _prefix("application-", app_name)


def storage(app_name: str) -> str:
    return _prefix("storage-", app_name)


def unit(unit_name: str) -> str:
    return _prefix("unit-", unit_name.replace("/", "-"))


def action(action_uuid: str) -> str:
    return _prefix("action-", action_uuid)


def space(space_name: str) -> str:
    return _prefix("space-", space_name)



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class BundleFacade(Type):
    name = "Bundle"
    version = 6

    @ReturnMapping(StringResult)
    async def ExportBundle(self, include_charm_defaults=None, include_series=None):
        """ExportBundle exports the current model configuration as bundle.

        include_charm_defaults : bool
        include_series : bool
        Returns -> StringResult
        """
        if include_charm_defaults is not None and not isinstance(
            include_charm_defaults, bool
        ):
            raise Exception(
                f"Expected include_charm_defaults to be a bool, received: {type(include_charm_defaults)}"
            )

        if include_series is not None and not isinstance(include_series, bool):
            raise Exception(
                f"Expected include_series to be a bool, received: {type(include_series)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Bundle", request="ExportBundle", version=6, params=_params)
        _params["include-charm-defaults"] = include_charm_defaults
        _params["include-series"] = include_series
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(BundleChangesResults)
    async def GetChanges(self, bundleurl=None, yaml=None):
        """GetChanges returns the list of changes required to deploy the given bundle
        data. The changes are sorted by requirements, so that they can be applied in
        order.
        GetChanges has been superseded in favour of GetChangesMapArgs. It's
        preferable to use that new method to add new functionality and move clients
        away from this one.

        bundleurl : str
        yaml : str
        Returns -> BundleChangesResults
        """
        if bundleurl is not None and not isinstance(bundleurl, (bytes, str)):
            raise Exception(
                f"Expected bundleurl to be a str, received: {type(bundleurl)}"
            )

        if yaml is not None and not isinstance(yaml, (bytes, str)):
            raise Exception(f"Expected yaml to be a str, received: {type(yaml)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Bundle", request="GetChanges", version=6, params=_params)
        _params["bundleURL"] = bundleurl
        _params["yaml"] = yaml
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(BundleChangesMapArgsResults)
    async def GetChangesMapArgs(self, bundleurl=None, yaml=None):
        """GetChangesMapArgs returns the list of changes required to deploy the given
        bundle data. The changes are sorted by requirements, so that they can be
        applied in order.
        V4 GetChangesMapArgs is not supported on anything less than v4

        bundleurl : str
        yaml : str
        Returns -> BundleChangesMapArgsResults
        """
        if bundleurl is not None and not isinstance(bundleurl, (bytes, str)):
            raise Exception(
                f"Expected bundleurl to be a str, received: {type(bundleurl)}"
            )

        if yaml is not None and not isinstance(yaml, (bytes, str)):
            raise Exception(f"Expected yaml to be a str, received: {type(yaml)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Bundle", request="GetChangesMapArgs", version=6, params=_params
        )
        _params["bundleURL"] = bundleurl
        _params["yaml"] = yaml
        reply = await self.rpc(msg)
        return reply


class CharmsFacade(Type):
    name = "Charms"
    version = 6

    @ReturnMapping(CharmOriginResult)
    async def AddCharm(self, charm_origin=None, force=None, url=None):
        """AddCharm adds the given charm URL (which must include revision) to the
        environment, if it does not exist yet. Local charms are not supported,
        only charm store and charm hub URLs. See also AddLocalCharm().

        charm_origin : CharmOrigin
        force : bool
        url : str
        Returns -> CharmOriginResult
        """
        if charm_origin is not None and not isinstance(
            charm_origin, (dict, CharmOrigin)
        ):
            raise Exception(
                f"Expected charm_origin to be a CharmOrigin, received: {type(charm_origin)}"
            )

        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if url is not None and not isinstance(url, (bytes, str)):
            raise Exception(f"Expected url to be a str, received: {type(url)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="AddCharm", version=6, params=_params)
        _params["charm-origin"] = charm_origin
        _params["force"] = force
        _params["url"] = url
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(Charm)
    async def CharmInfo(self, url=None):
        """CharmInfo returns information about the requested charm.

        url : str
        Returns -> Charm
        """
        if url is not None and not isinstance(url, (bytes, str)):
            raise Exception(f"Expected url to be a str, received: {type(url)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="CharmInfo", version=6, params=_params)
        _params["url"] = url
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def CheckCharmPlacement(self, placements=None):
        """CheckCharmPlacement checks if a charm is allowed to be placed with in a
        given application.

        placements : typing.Sequence[~ApplicationCharmPlacement]
        Returns -> ErrorResults
        """
        if placements is not None and not isinstance(placements, (bytes, str, list)):
            raise Exception(
                f"Expected placements to be a Sequence, received: {type(placements)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Charms", request="CheckCharmPlacement", version=6, params=_params
        )
        _params["placements"] = placements
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(DownloadInfoResults)
    async def GetDownloadInfos(self, entities=None):
        """GetDownloadInfos attempts to get the bundle corresponding to the charm url
        and origin.

        entities : typing.Sequence[~CharmURLAndOrigin]
        Returns -> DownloadInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="GetDownloadInfos", version=6, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(IsMeteredResult)
    async def IsMetered(self, url=None):
        """IsMetered returns whether or not the charm is metered.

        url : str
        Returns -> IsMeteredResult
        """
        if url is not None and not isinstance(url, (bytes, str)):
            raise Exception(f"Expected url to be a str, received: {type(url)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="IsMetered", version=6, params=_params)
        _params["url"] = url
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CharmsListResult)
    async def List(self, names=None):
        """List returns a list of charm URLs currently in the state.
        If supplied parameter contains any names, the result will
        be filtered to return only the charms with supplied names.

        names : typing.Sequence[str]
        Returns -> CharmsListResult
        """
        if names is not None and not isinstance(names, (bytes, str, list)):
            raise Exception(f"Expected names to be a Sequence, received: {type(names)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="List", version=6, params=_params)
        _params["names"] = names
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CharmResourcesResults)
    async def ListCharmResources(self, entities=None):
        """ListCharmResources returns a series of resources for a given charm.

        entities : typing.Sequence[~CharmURLAndOrigin]
        Returns -> CharmResourcesResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Charms", request="ListCharmResources", version=6, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ResolveCharmWithChannelResults)
    async def ResolveCharms(self, macaroon=None, resolve=None):
        """ResolveCharms resolves the given charm URLs with an optionally specified
        preferred channel.  Channel provided via CharmOrigin.

        macaroon : Macaroon
        resolve : typing.Sequence[~ResolveCharmWithChannel]
        Returns -> ResolveCharmWithChannelResults
        """
        if macaroon is not None and not isinstance(macaroon, (dict, Macaroon)):
            raise Exception(
                f"Expected macaroon to be a Macaroon, received: {type(macaroon)}"
            )

        if resolve is not None and not isinstance(resolve, (bytes, str, list)):
            raise Exception(
                f"Expected resolve to be a Sequence, received: {type(resolve)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Charms", request="ResolveCharms", version=6, params=_params)
        _params["macaroon"] = macaroon
        _params["resolve"] = resolve
        reply = await self.rpc(msg)
        return reply


class ClientFacade(Type):
    name = "Client"
    version = 6

    @ReturnMapping(FindToolsResult)
    async def FindTools(
        self, agentstream=None, arch=None, major=None, number=None, os_type=None
    ):
        """FindTools returns a List containing all tools matching the given parameters.
        TODO(juju 3.1) - remove, used by 2.9 client only

        agentstream : str
        arch : str
        major : int
        number : Number
        os_type : str
        Returns -> FindToolsResult
        """
        if agentstream is not None and not isinstance(agentstream, (bytes, str)):
            raise Exception(
                f"Expected agentstream to be a str, received: {type(agentstream)}"
            )

        if arch is not None and not isinstance(arch, (bytes, str)):
            raise Exception(f"Expected arch to be a str, received: {type(arch)}")

        if major is not None and not isinstance(major, int):
            raise Exception(f"Expected major to be a int, received: {type(major)}")

        if number is not None and not isinstance(number, (dict, Number)):
            raise Exception(f"Expected number to be a Number, received: {type(number)}")

        if os_type is not None and not isinstance(os_type, (bytes, str)):
            raise Exception(f"Expected os_type to be a str, received: {type(os_type)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="FindTools", version=6, params=_params)
        _params["agentstream"] = agentstream
        _params["arch"] = arch
        _params["major"] = major
        _params["number"] = number
        _params["os-type"] = os_type
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(FullStatus)
    async def FullStatus(self, patterns=None):
        """FullStatus gives the information needed for juju status over the api

        patterns : typing.Sequence[str]
        Returns -> FullStatus
        """
        if patterns is not None and not isinstance(patterns, (bytes, str, list)):
            raise Exception(
                f"Expected patterns to be a Sequence, received: {type(patterns)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="FullStatus", version=6, params=_params)
        _params["patterns"] = patterns
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StatusHistoryResults)
    async def StatusHistory(self, requests=None):
        """StatusHistory returns a slice of past statuses for several entities.

        requests : typing.Sequence[~StatusHistoryRequest]
        Returns -> StatusHistoryResults
        """
        if requests is not None and not isinstance(requests, (bytes, str, list)):
            raise Exception(
                f"Expected requests to be a Sequence, received: {type(requests)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="StatusHistory", version=6, params=_params)
        _params["requests"] = requests
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(AllWatcherId)
    async def WatchAll(self):
        """WatchAll initiates a watcher for entities in the connected model.

        Returns -> AllWatcherId
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Client", request="WatchAll", version=6, params=_params)

        reply = await self.rpc(msg)
        return reply


class SpacesFacade(Type):
    name = "Spaces"
    version = 6

    @ReturnMapping(ErrorResults)
    async def CreateSpaces(self, spaces=None):
        """CreateSpaces creates a new Juju network space, associating the
        specified subnets with it (optional; can be empty).

        spaces : typing.Sequence[~CreateSpaceParams]
        Returns -> ErrorResults
        """
        if spaces is not None and not isinstance(spaces, (bytes, str, list)):
            raise Exception(
                f"Expected spaces to be a Sequence, received: {type(spaces)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Spaces", request="CreateSpaces", version=6, params=_params)
        _params["spaces"] = spaces
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ListSpacesResults)
    async def ListSpaces(self):
        """ListSpaces lists all the available spaces and their associated subnets.

        Returns -> ListSpacesResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Spaces", request="ListSpaces", version=6, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(MoveSubnetsResults)
    async def MoveSubnets(self, args=None):
        """MoveSubnets ensures that the input subnets are in the input space.

        args : typing.Sequence[~MoveSubnetsParam]
        Returns -> MoveSubnetsResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Spaces", request="MoveSubnets", version=6, params=_params)
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def ReloadSpaces(self):
        """ReloadSpaces refreshes spaces from substrate

        Returns -> None
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Spaces", request="ReloadSpaces", version=6, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(RemoveSpaceResults)
    async def RemoveSpace(self, space_param=None):
        """RemoveSpace removes a space.
        Returns SpaceResults if entities/settings are found which makes the deletion not possible.

        space_param : typing.Sequence[~RemoveSpaceParam]
        Returns -> RemoveSpaceResults
        """
        if space_param is not None and not isinstance(space_param, (bytes, str, list)):
            raise Exception(
                f"Expected space_param to be a Sequence, received: {type(space_param)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Spaces", request="RemoveSpace", version=6, params=_params)
        _params["space-param"] = space_param
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def RenameSpace(self, changes=None):
        """RenameSpace renames a space.

        changes : typing.Sequence[~RenameSpaceParams]
        Returns -> ErrorResults
        """
        if changes is not None and not isinstance(changes, (bytes, str, list)):
            raise Exception(
                f"Expected changes to be a Sequence, received: {type(changes)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Spaces", request="RenameSpace", version=6, params=_params)
        _params["changes"] = changes
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ShowSpaceResults)
    async def ShowSpace(self, entities=None):
        """ShowSpace shows the spaces for a set of given entities.

        entities : typing.Sequence[~Entity]
        Returns -> ShowSpaceResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Spaces", request="ShowSpace", version=6, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply


class StorageFacade(Type):
    name = "Storage"
    version = 6

    @ReturnMapping(AddStorageResults)
    async def AddToUnit(self, storages=None):
        """AddToUnit validates and creates additional storage instances for units.
        A "CHANGE" block can block this operation.

        storages : typing.Sequence[~StorageAddParams]
        Returns -> AddStorageResults
        """
        if storages is not None and not isinstance(storages, (bytes, str, list)):
            raise Exception(
                f"Expected storages to be a Sequence, received: {type(storages)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="AddToUnit", version=6, params=_params)
        _params["storages"] = storages
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def Attach(self, ids=None):
        """Attach attaches existing storage instances to units.
        A "CHANGE" block can block this operation.

        ids : typing.Sequence[~StorageAttachmentId]
        Returns -> ErrorResults
        """
        if ids is not None and not isinstance(ids, (bytes, str, list)):
            raise Exception(f"Expected ids to be a Sequence, received: {type(ids)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="Attach", version=6, params=_params)
        _params["ids"] = ids
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def CreatePool(self, pools=None):
        """CreatePool creates a new pool with specified parameters.

        pools : typing.Sequence[~StoragePool]
        Returns -> ErrorResults
        """
        if pools is not None and not isinstance(pools, (bytes, str, list)):
            raise Exception(f"Expected pools to be a Sequence, received: {type(pools)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="CreatePool", version=6, params=_params)
        _params["pools"] = pools
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def DetachStorage(self, force=None, ids=None, max_wait=None):
        """DetachStorage sets the specified storage attachments to Dying, unless they are
        already Dying or Dead. Any associated, persistent storage will remain
        alive. This call can be forced.

        force : bool
        ids : StorageAttachmentIds
        max_wait : int
        Returns -> ErrorResults
        """
        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if ids is not None and not isinstance(ids, (dict, StorageAttachmentIds)):
            raise Exception(
                f"Expected ids to be a StorageAttachmentIds, received: {type(ids)}"
            )

        if max_wait is not None and not isinstance(max_wait, int):
            raise Exception(
                f"Expected max_wait to be a int, received: {type(max_wait)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="DetachStorage", version=6, params=_params)
        _params["force"] = force
        _params["ids"] = ids
        _params["max-wait"] = max_wait
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ImportStorageResults)
    async def Import(self, storage=None):
        """Import imports existing storage into the model.
        A "CHANGE" block can block this operation.

        storage : typing.Sequence[~ImportStorageParams]
        Returns -> ImportStorageResults
        """
        if storage is not None and not isinstance(storage, (bytes, str, list)):
            raise Exception(
                f"Expected storage to be a Sequence, received: {type(storage)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="Import", version=6, params=_params)
        _params["storage"] = storage
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(FilesystemDetailsListResults)
    async def ListFilesystems(self, filters=None):
        """ListFilesystems returns a list of filesystems in the environment matching
        the provided filter. Each result describes a filesystem in detail, including
        the filesystem's attachments.

        filters : typing.Sequence[~FilesystemFilter]
        Returns -> FilesystemDetailsListResults
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="ListFilesystems", version=6, params=_params)
        _params["filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StoragePoolsResults)
    async def ListPools(self, filters=None):
        """ListPools returns a list of pools.
        If filter is provided, returned list only contains pools that match
        the filter.
        Pools can be filtered on names and provider types.
        If both names and types are provided as filter,
        pools that match either are returned.
        This method lists union of pools and environment provider types.
        If no filter is provided, all pools are returned.

        filters : typing.Sequence[~StoragePoolFilter]
        Returns -> StoragePoolsResults
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="ListPools", version=6, params=_params)
        _params["filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StorageDetailsListResults)
    async def ListStorageDetails(self, filters=None):
        """ListStorageDetails returns storage matching a filter.

        filters : typing.Sequence[~StorageFilter]
        Returns -> StorageDetailsListResults
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Storage", request="ListStorageDetails", version=6, params=_params
        )
        _params["filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(VolumeDetailsListResults)
    async def ListVolumes(self, filters=None):
        """ListVolumes lists volumes with the given filters. Each filter produces
        an independent list of volumes, or an error if the filter is invalid
        or the volumes could not be listed.

        filters : typing.Sequence[~VolumeFilter]
        Returns -> VolumeDetailsListResults
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="ListVolumes", version=6, params=_params)
        _params["filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def Remove(self, storage=None):
        """Remove sets the specified storage entities to Dying, unless they are
        already Dying or Dead, such that the storage will eventually be removed
        from the model. If the arguments specify that the storage should be
        destroyed, then the associated cloud storage will be destroyed first;
        otherwise it will only be released from Juju's control.

        storage : typing.Sequence[~RemoveStorageInstance]
        Returns -> ErrorResults
        """
        if storage is not None and not isinstance(storage, (bytes, str, list)):
            raise Exception(
                f"Expected storage to be a Sequence, received: {type(storage)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="Remove", version=6, params=_params)
        _params["storage"] = storage
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def RemovePool(self, pools=None):
        """RemovePool deletes the named pool

        pools : typing.Sequence[~StoragePoolDeleteArg]
        Returns -> ErrorResults
        """
        if pools is not None and not isinstance(pools, (bytes, str, list)):
            raise Exception(f"Expected pools to be a Sequence, received: {type(pools)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="RemovePool", version=6, params=_params)
        _params["pools"] = pools
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StorageDetailsResults)
    async def StorageDetails(self, entities=None):
        """StorageDetails retrieves and returns detailed information about desired
        storage identified by supplied tags. If specified storage cannot be
        retrieved, individual error is returned instead of storage information.

        entities : typing.Sequence[~Entity]
        Returns -> StorageDetailsResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="StorageDetails", version=6, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def UpdatePool(self, pools=None):
        """UpdatePool deletes the named pool

        pools : typing.Sequence[~StoragePool]
        Returns -> ErrorResults
        """
        if pools is not None and not isinstance(pools, (bytes, str, list)):
            raise Exception(f"Expected pools to be a Sequence, received: {type(pools)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Storage", request="UpdatePool", version=6, params=_params)
        _params["pools"] = pools
        reply = await self.rpc(msg)
        return reply



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class AllModelWatcherFacade(Type):
    name = "AllModelWatcher"
    version = 4

    @ReturnMapping(AllWatcherNextResults)
    async def Next(self):
        """Next will return the current state of everything on the first call
        and subsequent calls will

        Returns -> AllWatcherNextResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="AllModelWatcher", request="Next", version=4, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def Stop(self):
        """Stop stops the watcher.

        Returns -> None
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="AllModelWatcher", request="Stop", version=4, params=_params)

        reply = await self.rpc(msg)
        return reply

    async def rpc(self, msg):
        """Patch rpc method to add Id."""
        if not hasattr(self, "Id"):
            raise RuntimeError('Missing "Id" field')
        msg["Id"] = id

        from .facade import TypeEncoder

        reply = await self.connection.rpc(msg, encoder=TypeEncoder)
        return reply


class ApplicationOffersFacade(Type):
    name = "ApplicationOffers"
    version = 4

    @ReturnMapping(ApplicationOffersResults)
    async def ApplicationOffers(self, bakery_version=None, offer_urls=None):
        """ApplicationOffers gets details about remote applications that match given URLs.

        bakery_version : int
        offer_urls : typing.Sequence[str]
        Returns -> ApplicationOffersResults
        """
        if bakery_version is not None and not isinstance(bakery_version, int):
            raise Exception(
                f"Expected bakery_version to be a int, received: {type(bakery_version)}"
            )

        if offer_urls is not None and not isinstance(offer_urls, (bytes, str, list)):
            raise Exception(
                f"Expected offer_urls to be a Sequence, received: {type(offer_urls)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="ApplicationOffers",
            version=4,
            params=_params,
        )
        _params["bakery-version"] = bakery_version
        _params["offer-urls"] = offer_urls
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def DestroyOffers(self, force=None, offer_urls=None):
        """DestroyOffers removes the offers specified by the given URLs, forcing if necessary.

        force : bool
        offer_urls : typing.Sequence[str]
        Returns -> ErrorResults
        """
        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if offer_urls is not None and not isinstance(offer_urls, (bytes, str, list)):
            raise Exception(
                f"Expected offer_urls to be a Sequence, received: {type(offer_urls)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers", request="DestroyOffers", version=4, params=_params
        )
        _params["force"] = force
        _params["offer-urls"] = offer_urls
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(QueryApplicationOffersResults)
    async def FindApplicationOffers(self, filters=None):
        """FindApplicationOffers gets details about remote applications that match given filter.

        filters : typing.Sequence[~OfferFilter]
        Returns -> QueryApplicationOffersResults
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="FindApplicationOffers",
            version=4,
            params=_params,
        )
        _params["Filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ConsumeOfferDetailsResults)
    async def GetConsumeDetails(self, offer_urls=None, user_tag=None):
        """GetConsumeDetails returns the details necessary to pass to another model
        to allow the specified args user to consume the offers represented by the args URLs.

        offer_urls : OfferURLs
        user_tag : str
        Returns -> ConsumeOfferDetailsResults
        """
        if offer_urls is not None and not isinstance(offer_urls, (dict, OfferURLs)):
            raise Exception(
                f"Expected offer_urls to be a OfferURLs, received: {type(offer_urls)}"
            )

        if user_tag is not None and not isinstance(user_tag, (bytes, str)):
            raise Exception(
                f"Expected user_tag to be a str, received: {type(user_tag)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="GetConsumeDetails",
            version=4,
            params=_params,
        )
        _params["offer-urls"] = offer_urls
        _params["user-tag"] = user_tag
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(QueryApplicationOffersResults)
    async def ListApplicationOffers(self, filters=None):
        """ListApplicationOffers gets deployed details about application offers that match given filter.
        The results contain details about the deployed applications such as connection count.

        filters : typing.Sequence[~OfferFilter]
        Returns -> QueryApplicationOffersResults
        """
        if filters is not None and not isinstance(filters, (bytes, str, list)):
            raise Exception(
                f"Expected filters to be a Sequence, received: {type(filters)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="ListApplicationOffers",
            version=4,
            params=_params,
        )
        _params["Filters"] = filters
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def ModifyOfferAccess(self, changes=None):
        """ModifyOfferAccess changes the application offer access granted to users.

        changes : typing.Sequence[~ModifyOfferAccess]
        Returns -> ErrorResults
        """
        if changes is not None and not isinstance(changes, (bytes, str, list)):
            raise Exception(
                f"Expected changes to be a Sequence, received: {type(changes)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="ModifyOfferAccess",
            version=4,
            params=_params,
        )
        _params["changes"] = changes
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def Offer(self, offers=None):
        """Offer makes application endpoints available for consumption at a specified URL.

        offers : typing.Sequence[~AddApplicationOffer]
        Returns -> ErrorResults
        """
        if offers is not None and not isinstance(offers, (bytes, str, list)):
            raise Exception(
                f"Expected offers to be a Sequence, received: {type(offers)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="ApplicationOffers", request="Offer", version=4, params=_params)
        _params["Offers"] = offers
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(RemoteApplicationInfoResults)
    async def RemoteApplicationInfo(self, bakery_version=None, offer_urls=None):
        """RemoteApplicationInfo returns information about the requested remote application.
        This call currently has no client side API, only there for the Dashboard at this stage.

        bakery_version : int
        offer_urls : typing.Sequence[str]
        Returns -> RemoteApplicationInfoResults
        """
        if bakery_version is not None and not isinstance(bakery_version, int):
            raise Exception(
                f"Expected bakery_version to be a int, received: {type(bakery_version)}"
            )

        if offer_urls is not None and not isinstance(offer_urls, (bytes, str, list)):
            raise Exception(
                f"Expected offer_urls to be a Sequence, received: {type(offer_urls)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ApplicationOffers",
            request="RemoteApplicationInfo",
            version=4,
            params=_params,
        )
        _params["bakery-version"] = bakery_version
        _params["offer-urls"] = offer_urls
        reply = await self.rpc(msg)
        return reply


class ModelGenerationFacade(Type):
    name = "ModelGeneration"
    version = 4

    @ReturnMapping(ErrorResult)
    async def AbortBranch(self, branch=None):
        """AbortBranch aborts the input branch, marking it complete.  However no
        changes are made applicable to the whole model.  No units may be assigned
        to the branch when aborting.

        branch : str
        Returns -> ErrorResult
        """
        if branch is not None and not isinstance(branch, (bytes, str)):
            raise Exception(f"Expected branch to be a str, received: {type(branch)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="AbortBranch", version=4, params=_params
        )
        _params["branch"] = branch
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResult)
    async def AddBranch(self, branch=None):
        """AddBranch adds a new branch with the input name to the model.

        branch : str
        Returns -> ErrorResult
        """
        if branch is not None and not isinstance(branch, (bytes, str)):
            raise Exception(f"Expected branch to be a str, received: {type(branch)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="AddBranch", version=4, params=_params
        )
        _params["branch"] = branch
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(BranchResults)
    async def BranchInfo(self, branches=None, detailed=None):
        """BranchInfo will return details of branch identified by the input argument,
        including units on the branch and the configuration disjoint with the
        master generation.
        An error is returned if no in-flight branch matching in input is found.

        branches : typing.Sequence[str]
        detailed : bool
        Returns -> BranchResults
        """
        if branches is not None and not isinstance(branches, (bytes, str, list)):
            raise Exception(
                f"Expected branches to be a Sequence, received: {type(branches)}"
            )

        if detailed is not None and not isinstance(detailed, bool):
            raise Exception(
                f"Expected detailed to be a bool, received: {type(detailed)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="BranchInfo", version=4, params=_params
        )
        _params["branches"] = branches
        _params["detailed"] = detailed
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(IntResult)
    async def CommitBranch(self, branch=None):
        """CommitBranch commits the input branch, making its changes applicable to
        the whole model and marking it complete.

        branch : str
        Returns -> IntResult
        """
        if branch is not None and not isinstance(branch, (bytes, str)):
            raise Exception(f"Expected branch to be a str, received: {type(branch)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="CommitBranch", version=4, params=_params
        )
        _params["branch"] = branch
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(BoolResult)
    async def HasActiveBranch(self, branch=None):
        """HasActiveBranch returns a true result if the input model has an "in-flight"
        branch matching the input name.

        branch : str
        Returns -> BoolResult
        """
        if branch is not None and not isinstance(branch, (bytes, str)):
            raise Exception(f"Expected branch to be a str, received: {type(branch)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="HasActiveBranch", version=4, params=_params
        )
        _params["branch"] = branch
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(BranchResults)
    async def ListCommits(self):
        """ListCommits will return the commits, hence only branches with generation_id higher than 0

        Returns -> BranchResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="ListCommits", version=4, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(GenerationResult)
    async def ShowCommit(self, generation_id=None):
        """ShowCommit will return details a commit given by its generationId
        An error is returned if either no branch can be found corresponding to the generation id.
        Or the generation id given is below 1.

        generation_id : int
        Returns -> GenerationResult
        """
        if generation_id is not None and not isinstance(generation_id, int):
            raise Exception(
                f"Expected generation_id to be a int, received: {type(generation_id)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="ShowCommit", version=4, params=_params
        )
        _params["generation-id"] = generation_id
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def TrackBranch(self, branch=None, entities=None, num_units=None):
        """TrackBranch marks the input units and/or applications as tracking the input
        branch, causing them to realise changes made under that branch.

        branch : str
        entities : typing.Sequence[~Entity]
        num_units : int
        Returns -> ErrorResults
        """
        if branch is not None and not isinstance(branch, (bytes, str)):
            raise Exception(f"Expected branch to be a str, received: {type(branch)}")

        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        if num_units is not None and not isinstance(num_units, int):
            raise Exception(
                f"Expected num_units to be a int, received: {type(num_units)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="ModelGeneration", request="TrackBranch", version=4, params=_params
        )
        _params["branch"] = branch
        _params["entities"] = entities
        _params["num-units"] = num_units
        reply = await self.rpc(msg)
        return reply


class SSHClientFacade(Type):
    name = "SSHClient"
    version = 4

    @ReturnMapping(SSHAddressesResults)
    async def AllAddresses(self, entities=None):
        """AllAddresses reports all addresses that might have SSH listening for each
        entity in args. The result is sorted with public addresses first.
        Machines and units are supported as entity types.

        entities : typing.Sequence[~Entity]
        Returns -> SSHAddressesResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="SSHClient", request="AllAddresses", version=4, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CloudSpecResult)
    async def ModelCredentialForSSH(self):
        """ModelCredentialForSSH returns a cloud spec for ssh purpose.
        This facade call is only used for k8s model.

        Returns -> CloudSpecResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="SSHClient", request="ModelCredentialForSSH", version=4, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(SSHAddressResults)
    async def PrivateAddress(self, entities=None):
        """PrivateAddress reports the preferred private network address for one or
        more entities. Machines and units are supported.

        entities : typing.Sequence[~Entity]
        Returns -> SSHAddressResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="SSHClient", request="PrivateAddress", version=4, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(SSHProxyResult)
    async def Proxy(self):
        """Proxy returns whether SSH connections should be proxied through the
        controller hosts for the model associated with the API connection.

        Returns -> SSHProxyResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="SSHClient", request="Proxy", version=4, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(SSHAddressResults)
    async def PublicAddress(self, entities=None):
        """PublicAddress reports the preferred public network address for one
        or more entities. Machines and units are supported.

        entities : typing.Sequence[~Entity]
        Returns -> SSHAddressResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="SSHClient", request="PublicAddress", version=4, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(SSHPublicKeysResults)
    async def PublicKeys(self, entities=None):
        """PublicKeys returns the public SSH hosts for one or more
        entities. Machines and units are supported.

        entities : typing.Sequence[~Entity]
        Returns -> SSHPublicKeysResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="SSHClient", request="PublicKeys", version=4, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example demonstrates how to obtain a formatted full status
description. For a similar solution using the FullStatus object
check examples/fullstatus.py
"""

import asyncio
import logging
import sys
import tempfile
from logging import getLogger

from juju.model import Model
from juju.status import formatted_status

LOG = getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


async def main():
    model = Model()
    await model.connect_current()

    application = await model.deploy(
        "ch:ubuntu",
        application_name="ubuntu",
        series="trusty",
        channel="stable",
    )

    await asyncio.sleep(10)
    tmp = tempfile.NamedTemporaryFile(delete=False)  # noqa: SIM115
    LOG.info("status dumped to %s", tmp.name)
    with open(tmp.name, "w") as f:
        for _ in range(10):
            # Uncomment this line to get the full status
            # using the standard output.
            # await formatted_status(model, target=sys.stdout)
            await formatted_status(model, target=f)
            f.write("-----------\n")
            await asyncio.sleep(1)
    await application.remove()
    await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
from __future__ import annotations

import asyncio
import base64
import os
import textwrap
import zipfile
from collections import defaultdict
from pathlib import Path
from typing import Any, Awaitable, Callable

import yaml
from pyasn1.codec.der.encoder import encode
from pyasn1.type import char, univ

from juju import _jasyncio

from . import errors, origin
from .client import client
from .errors import JujuError


async def execute_process(*cmd, log=None) -> bool:
    """Wrapper around asyncio.create_subprocess_exec."""
    p = await asyncio.create_subprocess_exec(
        *cmd,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await p.communicate()
    if log:
        log.debug("Exec %s -> %d", cmd, p.returncode)
        if stdout:
            log.debug(stdout.decode("utf-8"))
        if stderr:
            log.debug(stderr.decode("utf-8"))
    return p.returncode == 0


def juju_config_dir() -> str:
    """Resolves and returns the path string to the juju configuration folder
    for the juju CLI tool. Of the following items, returns the first option
    that works (top to bottom):

    * $JUJU_DATA
    * $XDG_DATA_HOME/juju
    * ~/.local/share/juju
    """
    # Set it to ~/.local/share/juju as default
    config_dir = Path("~/.local/share/juju")

    # Check $JUJU_DATA
    if juju_data := os.environ.get("JUJU_DATA"):
        config_dir = Path(juju_data)
    # Secondly check: $XDG_DATA_HOME for ~/.local/share
    elif xdg_data_home := os.environ.get("XDG_DATA_HOME"):
        config_dir = Path(xdg_data_home) / "juju"

    return str(config_dir.expanduser().resolve())


def juju_ssh_key_paths() -> tuple[str, str]:
    """Resolves and returns the path strings for public and private ssh keys
    for juju CLI.
    """
    config_dir = juju_config_dir()
    public_key_path = os.path.join(config_dir, "ssh", "juju_id_rsa.pub")
    private_key_path = os.path.join(config_dir, "ssh", "juju_id_rsa")

    return public_key_path, private_key_path


def _read_ssh_key() -> str:
    """Inner function for read_ssh_key, suitable for passing to our
    Executor.
    """
    public_key_path_str, _ = juju_ssh_key_paths()
    ssh_key_path = Path(public_key_path_str)
    with ssh_key_path.open("r") as ssh_key_file:
        ssh_key = ssh_key_file.readlines()[0].strip()
    return ssh_key


async def read_ssh_key() -> str:
    """Attempt to read the local juju admin's public ssh key, so that it can be
    passed on to a model.
    """
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _read_ssh_key)


class IdQueue:
    """Wrapper around asyncio.Queue that maintains a separate queue for each
    ID.
    """

    _queues: dict[int, asyncio.Queue[dict[str, Any] | Exception]]

    def __init__(self):
        self._queues = defaultdict(asyncio.Queue)
        # FIXME cleanup needed.
        # in some cases an Exception is put into the queue.
        # if the main coro exits, this exception will be logged as "never awaited"
        # we gotta do something about that to keep the output clean.
        #
        # Additionally, it's conceivable that a response is put in the queue
        # and then an exception is put via put_all()
        # the reader only ever fetches one item, and exception is "never awaited"
        # rewrite put_all to replace the pending response instead.

    async def get(self, id_: int) -> dict[str, Any]:
        value = await self._queues[id_].get()
        del self._queues[id_]
        if isinstance(value, Exception):
            raise value
        return value

    async def put(self, id_: int, value: dict[str, Any]):
        await self._queues[id_].put(value)

    async def put_all(self, value: Exception):
        for queue in self._queues.values():
            await queue.put(value)


async def block_until(
    *conditions: Callable[[], bool],
    timeout: float | None = None,
    wait_period: float = 0.5,
):
    """Return only after all conditions are true.

    If a timeout occurs, it cancels the task and raises
    asyncio.TimeoutError.
    """

    async def _block():
        while not all(c() for c in conditions):
            await asyncio.sleep(wait_period)

    await asyncio.shield(asyncio.wait_for(_block(), timeout))


async def block_until_with_coroutine(
    condition_coroutine: Callable[[], Awaitable[bool]],
    timeout: float | None = None,
    wait_period: float = 0.5,
):
    """Return only after the given coroutine returns True.

    If a timeout occurs, it cancels the task and raises
    asyncio.TimeoutError.
    """

    async def _block():
        while not await condition_coroutine():
            await asyncio.sleep(wait_period)

    await asyncio.shield(asyncio.wait_for(_block(), timeout=timeout))


async def wait_for_bundle(model, bundle: str | Path, **kwargs) -> None:
    """Helper to wait for just the apps in a specific bundle.

    Equivalent to loading the bundle, pulling out the app names, and calling::

        await model.wait_for_idle(app_names, **kwargs)
    """
    try:
        bundle_path = Path(bundle)
        if bundle_path.is_file():
            bundle = bundle_path.read_text()
        elif (bundle_path / "bundle.yaml").is_file():
            bundle = bundle_path / "bundle.yaml"
    except OSError:
        pass
    content: dict[str, Any] = yaml.safe_load(textwrap.dedent(str(bundle)).strip())
    apps = list(content.get("applications", content.get("services")).keys())
    await model.wait_for_idle(apps, **kwargs)


async def run_with_interrupt(task, *events: asyncio.Event, log=None):
    """Awaits a task while allowing it to be interrupted by one or more
    `asyncio.Event`s.

    If the task finishes without the events becoming set, the results of the
    task will be returned.  If the event become set, the task will be cancelled
    ``None`` will be returned.

    :param task: Task to run
    :param events: One or more `asyncio.Event`s which, if set, will interrupt
        `task` and cause it to be cancelled.
    """
    task = _jasyncio.create_task_with_handler(task, "tmp", log)
    event_tasks = [asyncio.ensure_future(event.wait()) for event in events]
    done, pending = await asyncio.wait(
        [task, *event_tasks], return_when=asyncio.FIRST_COMPLETED
    )
    for f in pending:
        f.cancel()  # cancel unfinished tasks
    for f in pending:
        try:
            await f
        except asyncio.CancelledError:
            pass
    for f in done:
        f.exception()  # prevent "exception was not retrieved" errors
    if task in done:
        return task.result()  # may raise exception
    else:
        return None


class Addrs(univ.SequenceOf):
    """Internal."""

    componentType = char.PrintableString()


class RegistrationInfo(univ.Sequence):
    """ASN.1 representation of:

    type RegistrationInfo struct { User string

    Addrs []string

    SecretKey []byte

        ControllerName string }
    """

    pass


def generate_user_controller_access_token(
    username: str, controller_endpoints, secret_key: str, controller_name
):
    """Implement in python what is currently done in GO.

    https://github.com/juju/juju/blob/a5ab92e/cmd/juju/user/utils.go#L16

    :param username: name of the user to register
    :param controller_endpoints: juju controller endpoints list in the format <ip>:<port>
    :param secret_key: base64 encoded string of the secret-key generated by juju
    :param controller_name: name of the controller to register to.
    """
    # Secret key is returned as base64 encoded string in:
    # https://websockets.readthedocs.io/en/stable/_modules/websockets/protocol.html#WebSocketCommonProtocol.recv
    # Decoding it before marshalling into the ASN.1 message
    secret_key = base64.b64decode(secret_key)
    addr = Addrs()
    for endpoint in controller_endpoints:
        addr.append(endpoint)

    registration_string = RegistrationInfo()
    registration_string.setComponentByPosition(0, char.PrintableString(username))
    registration_string.setComponentByPosition(1, addr)
    registration_string.setComponentByPosition(2, univ.OctetString(secret_key))
    registration_string.setComponentByPosition(3, char.PrintableString(controller_name))
    registration_string = encode(registration_string)
    remainder = len(registration_string) % 3
    registration_string += b"\0" * (3 - remainder)
    return base64.urlsafe_b64encode(registration_string)


def get_local_charm_data(path: str | Path, yaml_file: str) -> dict[str, Any]:
    """Retrieve Metadata of a Charm from its path.

    :param str path: Path of charm directory or .charm file
    :param str yaml_file: name of the yaml file, can be either
    "metadata.yaml", or "manifest.yaml", or "charmcraft.yaml"

    :return: Object of charm metadata
    """
    if str(path).endswith(".charm"):
        with zipfile.ZipFile(str(path), "r") as charm_file:
            metadata = yaml.load(charm_file.read(yaml_file), Loader=yaml.FullLoader)
    else:
        entity_path = Path(path)
        metadata_path = entity_path / yaml_file
        if not metadata_path.exists():
            return {}
        metadata = yaml.load(metadata_path.read_text(), Loader=yaml.FullLoader)

    return metadata


def get_local_charm_metadata(path: str | Path) -> dict[str, Any]:
    return get_local_charm_data(path, "metadata.yaml")


def get_local_charm_manifest(path: str | Path) -> dict[str, Any]:
    return get_local_charm_data(path, "manifest.yaml")


def get_local_charm_charmcraft_yaml(path: str | Path) -> dict[str, Any]:
    return get_local_charm_data(path, "charmcraft.yaml")


PRECISE = "precise"
QUANTAL = "quantal"
RARING = "raring"
SAUCY = "saucy"
TRUSTY = "trusty"
UTOPIC = "utopic"
VIVID = "vivid"
WILY = "wily"
XENIAL = "xenial"
YAKKETY = "yakkety"
ZESTY = "zesty"
ARTFUL = "artful"
BIONIC = "bionic"
COSMIC = "cosmic"
DISCO = "disco"
EOAN = "eoan"
FOCAL = "focal"
GROOVY = "groovy"
HIRSUTE = "hirsute"
IMPISH = "impish"
JAMMY = "jammy"
KINETIC = "kinetic"
LUNAR = "lunar"
MANTIC = "mantic"
NOBLE = "noble"

UBUNTU_SERIES = {
    PRECISE: "12.04",
    QUANTAL: "12.10",
    RARING: "13.04",
    SAUCY: "13.10",
    TRUSTY: "14.04",
    UTOPIC: "14.10",
    VIVID: "15.04",
    WILY: "15.10",
    XENIAL: "16.04",
    YAKKETY: "16.10",
    ZESTY: "17.04",
    ARTFUL: "17.10",
    BIONIC: "18.04",
    COSMIC: "18.10",
    DISCO: "19.04",
    EOAN: "19.10",
    FOCAL: "20.04",
    GROOVY: "20.10",
    HIRSUTE: "21.04",
    IMPISH: "21.10",
    JAMMY: "22.04",
    KINETIC: "22.10",
    LUNAR: "23.04",
    MANTIC: "23.10",
    NOBLE: "24.04",
}

KUBERNETES = "kubernetes"
KUBERNETES_SERIES = {KUBERNETES: "kubernetes"}

ALL_SERIES_VERSIONS = {**UBUNTU_SERIES, **KUBERNETES_SERIES}


def get_series_version(series_name: str) -> str:
    """get_series_version outputs the version of the OS based on the given
    series e.g. jammy -> 22.04, kubernetes -> kubernetes.

    :param str series_name: name of the series
    :return str: os version
    """
    if series_name not in ALL_SERIES_VERSIONS:
        raise errors.JujuError("Unknown series : %s", series_name)
    return ALL_SERIES_VERSIONS[series_name]


def get_version_series(version: str) -> str:
    """get_version_series is the opposite of the get_series_version. It outputs
    the series based on given OS version.

    :param str version: version of the OS
    return str: name of the series corresponding to the given version
    """
    if version not in UBUNTU_SERIES.values():
        raise errors.JujuError("Unknown version : %s", version)
    return list(UBUNTU_SERIES.keys())[list(UBUNTU_SERIES.values()).index(version)]


def get_local_charm_base(series: str, charm_path: str, base_class: type):
    """Deduce the base [channel/osname] of a local charm based on what we know
    already.

    :param str series: This may come from the argument or the
        metadata.yaml
    :param str charm_path: Path of charm directory/.charm file
    :param class base_class:
    :return: Instance of the baseCls with channel/osname information
    """
    channel_for_base = ""
    os_name_for_base = ""

    # We should know the series, so use it to get a channel
    if series:
        channel_for_base = get_series_version(series) if series else ""
        if channel_for_base:
            # we currently only support ubuntu series (statically)
            # TODO (cderici) : go juju/core/series/supported.go and get the
            #  others here too
            if series in KUBERNETES_SERIES:
                os_name_for_base = "kubernetes"
            else:
                os_name_for_base = "ubuntu"

    # Check the charm manifest
    if channel_for_base == "":
        charm_manifest = get_local_charm_manifest(charm_path)
        if "bases" in charm_manifest:
            channel_for_base = charm_manifest["bases"][0]["channel"]
            os_name_for_base = charm_manifest["bases"][0]["name"]

    # Also check the charmcraft.yaml
    if channel_for_base == "":
        charmcraft_yaml = get_local_charm_charmcraft_yaml(charm_path)
        if "bases" in charmcraft_yaml:
            channel_for_base = charmcraft_yaml["bases"][0]["run-on"][0]["channel"]
            os_name_for_base = charmcraft_yaml["bases"][0]["run-on"][0]["name"]

    if channel_for_base == "":
        raise errors.JujuError("Unable to determine base for charm : %s" % charm_path)

    # Legacy k8s charms - assume ubuntu focal
    # as per juju/cmd/juju/application/utils.DeduceOrigin()
    if channel_for_base == "kubernetes" or os_name_for_base == "kubernetes":
        channel_for_base = "20.04/stable"
        os_name_for_base = "ubuntu"
    return base_class(channel_for_base, os_name_for_base)


def base_channel_to_series(channel: str) -> str:
    """Returns the series string using the track inside the base channel.

    :param str channel: is track/risk (e.g. 20.04/stable)
    :return: str series (e.g. focal)
    """
    return get_version_series(origin.Channel.parse(channel).track)


def parse_base_arg(base: str) -> client.Base:
    """Parses a given base into a Client.Base object :param base str : The base
    to deploy a charm (e.g. ubuntu@22.04)
    """
    client.CharmBase()
    if not (isinstance(base, str) and "@" in base):
        raise errors.JujuError(
            f"expected base string to contain os and channel separated by '@', got : {base}"
        )

    name, channel = base.split("@")
    return client.Base(name=name, channel=channel)


DEFAULT_SUPPORTED_LTS = "jammy"
DEFAULT_SUPPORTED_LTS_BASE = client.Base(channel="22.04", name="ubuntu")


def base_channel_from_series(track, risk, series):
    return (
        origin.Channel(track=track, risk=risk)
        .normalize()
        .compute_base_channel(series=series)
    )


def get_os_from_series(series: str) -> str:
    if series in UBUNTU_SERIES:
        return "ubuntu"
    raise JujuError(f"os for the series {series} needs to be added")


def get_base_from_origin_or_channel(origin_or_channel, series=None):
    channel, os_name = None, None
    if series:
        channel = base_channel_from_series(
            origin_or_channel.track, origin_or_channel.risk, series
        )
        os_name = get_os_from_series(series)
    return client.Base(channel=channel, name=os_name)


def series_for_charm(requested_series: str, supported_series: list[str]) -> str:
    """series_for_charm takes a requested series and a list of series supported
    by a charm and returns the series which is relevant.

    If the requested series is empty, then the first supported series is
    used, otherwise the requested series is validated against the
    supported series.
    """
    if len(supported_series) == 1 and supported_series[0] == "":
        raise JujuError("invalid supported series reported by charm : ['']")
    if len(supported_series) == 0:
        if requested_series == "":
            raise JujuError("missing series")
        return requested_series

    # use the charm default
    if requested_series == "":
        return supported_series[-1]

    for s in supported_series:
        if requested_series == s:
            return requested_series
    raise JujuError(
        f"requested series {requested_series} is not among the supported series {supported_series}"
    )


def user_requested(series_arg: str, supported_series: list[str], force: bool) -> str:
    series = series_for_charm(series_arg, supported_series)
    if force:
        series = series_arg
    # Todo (cderici): validate the series with workload_series to see if juju is
    # supporting that
    return series


def series_selector(
    series_arg: str = "",
    charm_url=None,
    model_config=None,
    supported_series: list[str] = [],
    force: bool = False,
) -> str:
    """Select series to deploy on.

    series_selector corresponds to the CharmSeries() in
    https://github.com/juju/juju/blob/develop/core/charm/series_selector.go

    determines what series to use with a charm.
    Order of preference is:
    - user requested with --series or defined by bundle when deploying
    - user requested in charm's url (e.g. juju deploy jammy/ubuntu)
    - model default, if set, acts like --series
    - default from charm metadata supported series / series in url
    - default LTS
    """
    # User has requested a series with --series.
    if series_arg:
        return user_requested(series_arg, supported_series, force)

    # User specified a series in the charm URL, e.g.
    # juju deploy precise/ubuntu.
    if charm_url and charm_url.series:
        return user_requested(charm_url.series, supported_series, force)

    # No series explicitly requested by the user.
    # Use model default series, if explicitly set and supported by the charm.
    if model_config and model_config["default-base"].value:
        default_base = model_config["default-base"].value
        base = parse_base_arg(default_base)
        series = base_channel_to_series(base.channel)
        return user_requested(series, supported_series, force)

    # Next fall back to the charm's list of series, filtered to what's supported
    # by Juju. Preserve the order of the supported series from the charm
    # metadata, as the order could be out of order compared to Ubuntu series
    # order (precise, xenial, bionic, trusty, etc).
    try:
        # TODO (cderici): restrict the supported_series with JujuSupportedSeries
        return user_requested("", supported_series, force)
    except JujuError:
        pass

    # Charm hasn't specified a default (likely due to being a local charm
    # deployed by path). Last chance, best we can do is default to LTS.
    return DEFAULT_SUPPORTED_LTS


def should_upgrade_resource(
    available_resource: dict[str, str], existing_resources, arg_resources
) -> bool:
    """Determine if the given resource should be upgraded.

    Called in the context of upgrade_charm. Given a resource R, takes a look
    at the resources we already have and decides if we need to refresh R.

    :param dict[str] available_resource: The dict representing the
    client.Resource coming from the charmhub api. We're considering if
    we need to refresh this during upgrade_charm.
    :param dict[str] existing_resources: The dict coming from
    resources_facade.ListResources representing the resources of the
    currently deployed charm.
    :param dict[str] arg_resources: user provided resources to be refreshed

    :result bool: The decision to refresh the given resource
    """
    # should upgrade resource?
    res_name = available_resource.get("Name", available_resource.get("name"))

    if res_name in arg_resources:
        return True

    # do we have it already?
    if res_name in existing_resources:
        # no upgrade, if it's upload
        if existing_resources[res_name].origin == "upload":
            return False
        # no upgrade, if upstream doesn't have a newer revision of the resource
        # available
        available_rev = available_resource.get(
            "Revision", available_resource.get("revision", -1)
        )
        u_fields = existing_resources[res_name].unknown_fields
        existing_rev = u_fields.get("Revision", u_fields.get("revision", -1))
        if existing_rev >= available_rev:
            return False
    return True



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio

from juju.model import Model


async def main():
    charm = "juju-qa-test"

    model = Model()
    print("Connecting to model")
    # connect to current model with current user, per Juju CLI
    await model.connect()

    try:
        print(f"Deploying {charm} --channel 2.0/stable --revision 22")
        application = await model.deploy(
            "juju-qa-test",
            application_name="test",
            channel="2.0/stable",
            revision=22,
        )

        print("Waiting for active")
        await model.wait_for_idle(status="active")

        print(f"Removing {charm}")
        await application.remove()
    finally:
        print("Disconnecting from model")
        await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio
import json

import requests

from .client import client
from .errors import JujuError


class CharmHub:
    def __init__(self, model):
        self.model = model

    async def _charmhub_url(self):
        model_conf = await self.model.get_config()
        return model_conf["charmhub-url"]

    async def request_charmhub_with_retry(self, url, retries):
        for _ in range(retries):
            _response = requests.get(url)  # noqa: S113
            if _response.status_code == 200:
                return _response
            await asyncio.sleep(5)
        raise JujuError(f"Got {_response.status_code} from {url}")

    async def get_charm_id(self, charm_name):
        _conn, _headers, _path_prefix = self.model.connection().https_connection()

        charmhub_url = await self._charmhub_url()
        url = f"{charmhub_url.value}/v2/charms/info/{charm_name}"
        _response = await self.request_charmhub_with_retry(url, 5)
        response = json.loads(_response.text)
        return response["id"], response["name"]

    async def is_subordinate(self, charm_name):
        _conn, _headers, _path_prefix = self.model.connection().https_connection()

        charmhub_url = await self._charmhub_url()
        url = f"{charmhub_url.value}/v2/charms/info/{charm_name}?fields=default-release.revision.subordinate"
        _response = await self.request_charmhub_with_retry(url, 5)
        response = json.loads(_response.text)
        rev_response = response["default-release"]["revision"]
        return "subordinate" in rev_response and rev_response["subordinate"]

    # TODO (caner) : we should be able to recreate the channel-map through the
    #  api call without needing the CharmHub facade

    async def list_resources(self, charm_name):
        _conn, _headers, __path_prefix = self.model.connection().https_connection()

        charmhub_url = await self._charmhub_url()
        url = f"{charmhub_url.value}/v2/charms/info/{charm_name}?fields=default-release.resources"
        _response = await self.request_charmhub_with_retry(url, 5)
        response = json.loads(_response.text)
        return response["default-release"]["resources"]

    async def info(self, name, channel=None):
        """Info displays detailed information about a CharmHub charm. The charm
        can be specified by the exact name.

        Channel is a hint for providing the metadata for a given channel.
        Without the channel hint then only the default release will have the
        metadata.

        """
        if not name:
            raise JujuError("name expected")

        if self.model.connection().is_using_old_client:
            if channel is None:
                channel = ""
            facade = self._facade()
            res = await facade.Info(tag=f"application-{name}", channel=channel)
            err_code = res.errors.error_list.code
            if err_code:
                raise JujuError(
                    f"charmhub.info - {err_code} : {res.errors.error_list.message}"
                )
            result = res.result
            result.channel_map = CharmHub._channel_map_to_dict(
                result.channel_map, name, channel=channel
            )
            result = result.serialize()
        else:
            charmhub_url = await self._charmhub_url()
            url = f"{charmhub_url.value}/v2/charms/info/{name}?fields=channel-map"
            try:
                _response = await self.request_charmhub_with_retry(url, 5)
            except JujuError as e:
                if "404" in e.message:
                    raise JujuError(f"{name} not found") from e
            result = json.loads(_response.text)
            result["channel-map"] = CharmHub._channel_list_to_map(
                result["channel-map"], name, channel=channel
            )
        return result

    @staticmethod
    def _channel_list_to_map(channel_list_map, name, channel=""):
        """Charmhub API returns the channel map as a list of channel objects
        (with risk, track, revision, download etc). This turns that into a map
        that's keyed with the channel=track/risk for easy
        filtering/processing. This representation is also closer to the
        result of the 2.9 facade call.

        So basically,
        [{'channel':{'risk':'stable', 'track':'latest'}, 'revision': 58}]
          becomes:
        {'latest/stable': {'channel':{'risk':'stable', 'track':'latest'},
        'revision': 58}}

        :param channel_list_map: [map[str][any]]
        :return: map[str][map[str][any]]
        """
        channel_map = {}
        for ch in channel_list_map:
            ch_name = f"{ch['channel']['track']}/{ch['channel']['risk']}"
            if channel and channel != ch_name:
                # If channel is given, then filter out the rest
                continue
            channel_map[ch_name] = ch
            if channel == ch_name:
                # If we found the desired channel, no need to continue
                break
        # After loop is done, check for non-existent channel
        if channel and channel not in channel_map:
            raise JujuError(f"Charmhub.info : channel {channel} not found for {name}")
        return channel_map

    @staticmethod
    def _channel_map_to_dict(channel_map, name, channel=""):
        """Converts the client.definitions.Channel objects into python maps
        inside a channel map (for pylibjuju <3.0)

        :param channel_map: map[str][Channel]
        :return: map[str][map[str][any]]
        """
        channel_dict = {}
        for ch_name, ch_obj in channel_map.items():
            # No need to worry about filtering channel
            # Charmhub facade will take care of that
            _ch = ch_obj.serialize()
            _ch["platforms"] = [p.serialize() for p in _ch["platforms"]]
            channel_dict[ch_name] = _ch
        if channel and channel not in channel_dict:
            raise JujuError(f"Charmhub.info : channel {channel} not found for {name}")
        return channel_dict

    async def find(
        self,
        query,
        category=None,
        channel=None,
        charm_type=None,
        platforms=None,
        publisher=None,
        relation_requires=None,
        relation_provides=None,
    ):
        """Find queries the CharmHub store for available charms or bundles."""
        if charm_type is not None and charm_type not in ["charm", "bundle"]:
            raise JujuError("expected either charm or bundle for charm_type")

        facade = self._facade()
        return await facade.Find(
            query=query,
            category=category,
            channel=channel,
            type_=charm_type,
            platforms=platforms,
            publisher=publisher,
            relation_provides=relation_provides,
            relation_requires=relation_requires,
        )

    def _facade(self):
        return client.CharmHubFacade.from_connection(self.model.connection())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Deploy a bundle and waits until it reports itself active
3. Destroys the units and applications

"""

import asyncio

from juju.model import Model


async def main():
    model = Model()
    print("Connecting to model")
    # Connect to current model with current user, per Juju CLI
    await model.connect()

    try:
        print("Deploying bundle")
        await model.deploy(
            "./examples/k8s-local-bundle/big-k8s-bundle.yaml",
            channel="edge",
            trust=True,
        )
    finally:
        print("Disconnecting from model")
        await model.disconnect()
        print("Success")


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio
import ipaddress
import logging
import typing

from backports.datetime_fromisoformat import datetime_fromisoformat

from juju.utils import block_until, juju_ssh_key_paths

from . import model, tag
from .annotationhelper import _get_annotations, _set_annotations
from .client import client
from .errors import JujuError

log = logging.getLogger(__name__)


class Machine(model.ModelEntity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def destroy(self, force=False):
        """Remove this machine from the model.

        Blocks until the machine is actually removed.

        """
        if self.connection.is_using_old_client:
            # Then we'll use the DestroyMachines from client.ClientFacade
            facade = client.ClientFacade.from_connection(self.connection)
            await facade.DestroyMachines(force=force, machine_names=[self.id])
        else:
            facade = client.MachineManagerFacade.from_connection(self.connection)
            await facade.DestroyMachineWithParams(
                force=force, machine_tags=[tag.machine(self.id)]
            )

        log.debug("Destroying machine %s", self.id)
        return await self.model._wait("machine", self.id, "remove")

    remove = destroy

    async def get_annotations(self):
        """Get annotations on this machine.

        :return dict: The annotations for this application
        """
        return await _get_annotations(self.tag, self.connection)

    async def set_annotations(self, annotations):
        """Set annotations on this machine.

        :param annotations map[string]string: the annotations as key/value
            pairs.

        """
        return await _set_annotations(self.tag, annotations, self.connection)

    def _format_addr(self, addr):
        """Validate and format IP address.

        :param addr: IPv6 or IPv4 address
        :type addr: str
        :returns: Address string, optionally encapsulated in brackets ([])
        :rtype: str
        :raises: ValueError
        """
        ipaddr = ipaddress.ip_address(addr)
        if isinstance(ipaddr, ipaddress.IPv6Address):
            fmt = "[{}]"
        else:
            fmt = "{}"
        return fmt.format(ipaddr)

    async def scp_to(
        self,
        source,
        destination,
        user="ubuntu",
        proxy=False,
        scp_opts="",
        wait_for_active=False,
        timeout=None,
    ):
        """Transfer files to this machine.

        :param str source: Local path of file(s) to transfer
        :param str destination: Remote destination of transferred files
        :param str user: Remote username
        :param bool proxy: Proxy through the Juju API server
        :param scp_opts: Additional options to the `scp` command
        :type scp_opts: str or list
        :param bool wait_for_active: Wait until the machine is ready to take in ssh commands.
        :param int timeout: Time in seconds to wait until the machine becomes ready.
        """
        if proxy:
            raise NotImplementedError("proxy option is not implemented")
        if wait_for_active:
            await block_until(lambda: self.addresses, timeout=timeout)
        try:
            # if dns_name is an IP address format it appropriately
            address = self._format_addr(self.dns_name)
        except ValueError:
            # otherwise we assume it to be a DNS resolvable string
            address = self.dns_name
        destination = f"{user}@{address}:{destination}"
        await self._scp(source, destination, scp_opts)

    async def scp_from(
        self,
        source,
        destination,
        user="ubuntu",
        proxy=False,
        scp_opts="",
        wait_for_active=False,
        timeout=None,
    ):
        """Transfer files from this machine.

        :param str source: Remote path of file(s) to transfer
        :param str destination: Local destination of transferred files
        :param str user: Remote username
        :param bool proxy: Proxy through the Juju API server
        :param scp_opts: Additional options to the `scp` command
        :type scp_opts: str or list
        :param bool wait_for_active: Wait until the machine is ready to take in ssh commands.
        :param int timeout: Time in seconds to wait until the machine becomes ready.
        """
        if proxy:
            raise NotImplementedError("proxy option is not implemented")
        if wait_for_active:
            await block_until(lambda: self.addresses, timeout=timeout)
        try:
            # if dns_name is an IP address format it appropriately
            address = self._format_addr(self.dns_name)
        except ValueError:
            # otherwise we assume it to be a DNS resolvable string
            address = self.dns_name
        source = f"{user}@{address}:{source}"
        await self._scp(source, destination, scp_opts)

    async def _scp(self, source, destination, scp_opts):
        """Execute an scp command. Requires a fully qualified source and
        destination.
        """
        _, id_path = juju_ssh_key_paths()
        cmd = ["scp", "-i", id_path, "-o", "StrictHostKeyChecking=no", "-q", "-B"]
        cmd.extend(scp_opts.split() if isinstance(scp_opts, str) else scp_opts)
        cmd.extend([source, destination])
        # There's a bit of a gap between the time that the machine is assigned an IP and the ssh
        # service is up and listening, which creates a race for the ssh command. So we retry a
        # couple of times until either we run out of attempts, or the ssh command succeeds to
        # mitigate that effect.
        # TODO (cderici): refactor the ssh and scp subcommand processing into a single method.
        retry_backoff = 2
        retries = 10
        for _ in range(retries):
            process = await asyncio.create_subprocess_exec(*cmd)
            await process.wait()
            if process.returncode == 0:
                break
            await asyncio.sleep(retry_backoff)
        if process.returncode != 0:
            raise JujuError(f"command failed after {retries} attempts: {cmd}")

    async def ssh(
        self,
        command,
        user="ubuntu",
        proxy=False,
        ssh_opts=None,
        wait_for_active=False,
        timeout=None,
    ):
        """Execute a command over SSH on this machine.

        :param str command: Command to execute
        :param str user: Remote username
        :param bool proxy: Proxy through the Juju API server
        :param str ssh_opts: Additional options to the `ssh` command
        :param bool wait_for_active: Wait until the machine is ready to take in ssh commands.
        :param int timeout: Time in seconds to wait until the machine becomes ready.
        """
        if proxy:
            raise NotImplementedError("proxy option is not implemented")
        if wait_for_active:
            await block_until(lambda: self.addresses, timeout=timeout)
        address = self.dns_name
        destination = f"{user}@{address}"
        _, id_path = juju_ssh_key_paths()
        cmd = [
            "ssh",
            "-i",
            id_path,
            "-o",
            "StrictHostKeyChecking=no",
            "-q",
            destination,
        ]
        if ssh_opts:
            cmd.extend(ssh_opts.split() if isinstance(ssh_opts, str) else ssh_opts)
        cmd.extend([command])

        # There's a bit of a gap between the time that the machine is assigned an IP and the ssh
        # service is up and listening, which creates a race for the ssh command. So we retry a
        # couple of times until either we run out of attempts, or the ssh command succeeds to
        # mitigate that effect.
        retry_backoff = 2
        retries = 10
        for _ in range(retries):
            process = await asyncio.create_subprocess_exec(
                *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            stdout, stderr = await process.communicate()
            if process.returncode == 0:
                break
            await asyncio.sleep(retry_backoff)
        if process.returncode != 0:
            raise JujuError(
                f"command failed: {cmd} after {retries} attempts, with {stderr.decode()}"
            )
        # stdout is a bytes-like object, returning a string might be more useful
        return stdout.decode()

    @property
    def addresses(self) -> typing.List[dict]:
        """Returns the machine addresses."""
        return self.safe_data["addresses"] or []

    @property
    def agent_status(self):
        """Returns the current Juju agent status string."""
        return self.safe_data["agent-status"]["current"]

    @property
    def agent_status_since(self):
        """Get the time when the `agent_status` was last updated."""
        return datetime_fromisoformat(self.safe_data["agent-status"]["since"])

    @property
    def agent_version(self):
        """Get the version of the Juju machine agent.

        May return None if the agent is not yet available.
        """
        version = self.safe_data["agent-status"]["version"]
        if version:
            return client.Number.from_json(version)
        else:
            return None

    @property
    def status(self):
        """Returns the current machine provisioning status string."""
        return self.safe_data["instance-status"]["current"]

    @property
    def status_message(self):
        """Returns the current machine provisioning status message."""
        return self.safe_data["instance-status"]["message"]

    @property
    def status_since(self):
        """Get the time when the `status` was last updated."""
        return datetime_fromisoformat(self.safe_data["instance-status"]["since"])

    @property
    def dns_name(self):
        """Get the DNS name for this machine. This is a best guess based on the
        addresses available in current data.

        May return None if no suitable address is found.
        """
        ordered_scopes = ["public", "local-cloud", "local-fan"]
        ordered_addresses = [
            address
            for scope in ordered_scopes
            for address in self.addresses
            if scope == address["scope"]
        ]

        for address in ordered_addresses:
            scope = address["scope"]
            for check_scope in ordered_scopes:
                if scope == check_scope:
                    return address["value"]
        return None

    @property
    def hostname(self):
        """Get the hostname for this machine as reported by the machine agent
        running on it. This is only supported on 2.8.10+ controllers.

        May return None if no hostname information is available.
        """
        if "hostname" in self.safe_data and self.safe_data["hostname"] != "":
            return self.safe_data["hostname"]
        return None

    @property
    def series(self):
        """Returns the series of the current machine"""
        return self.safe_data["series"]

    @property
    def tag(self):
        return tag.machine(self.id)



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class ControllerFacade(Type):
    name = "Controller"
    version = 12

    @ReturnMapping(UserModelList)
    async def AllModels(self):
        """AllModels allows controller administrators to get the list of all the
        models in the controller.

        Returns -> UserModelList
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Controller", request="AllModels", version=12, params=_params)

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CloudSpecResults)
    async def CloudSpec(self, entities=None):
        """CloudSpec returns the model's cloud spec.

        entities : typing.Sequence[~Entity]
        Returns -> CloudSpecResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Controller", request="CloudSpec", version=12, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def ConfigSet(self, config=None):
        """ConfigSet changes the value of specified controller configuration
        settings. Only some settings can be changed after bootstrap.
        Settings that aren't specified in the params are left unchanged.

        config : typing.Mapping[str, typing.Any]
        Returns -> None
        """
        if config is not None and not isinstance(config, dict):
            raise Exception(
                f"Expected config to be a Mapping, received: {type(config)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Controller", request="ConfigSet", version=12, params=_params)
        _params["config"] = config
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ControllerAPIInfoResults)
    async def ControllerAPIInfoForModels(self, entities=None):
        """ControllerAPIInfoForModels returns the controller api connection details for the specified models.

        entities : typing.Sequence[~Entity]
        Returns -> ControllerAPIInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller",
            request="ControllerAPIInfoForModels",
            version=12,
            params=_params,
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ControllerConfigResult)
    async def ControllerConfig(self):
        """ControllerConfig returns the controller's configuration.

        Returns -> ControllerConfigResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="ControllerConfig", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ControllerVersionResults)
    async def ControllerVersion(self):
        """ControllerVersion returns the version information associated with this
        controller binary.

        NOTE: the implementation intentionally does not check for SuperuserAccess
        as the Version is known even to users with login access.

        Returns -> ControllerVersionResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="ControllerVersion", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(DashboardConnectionInfo)
    async def DashboardConnectionInfo(self):
        """DashboardConnectionInfo returns the connection information for a client to
        connect to the Juju Dashboard including any proxying information.

        Returns -> DashboardConnectionInfo
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller",
            request="DashboardConnectionInfo",
            version=12,
            params=_params,
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def DestroyController(
        self,
        destroy_models=None,
        destroy_storage=None,
        force=None,
        max_wait=None,
        model_timeout=None,
    ):
        """DestroyController destroys the controller.

        If the args specify the destruction of the models, this method will
        attempt to do so. Otherwise, if the controller has any non-empty,
        non-Dead hosted models, then an error with the code
        params.CodeHasHostedModels will be transmitted.

        destroy_models : bool
        destroy_storage : bool
        force : bool
        max_wait : int
        model_timeout : int
        Returns -> None
        """
        if destroy_models is not None and not isinstance(destroy_models, bool):
            raise Exception(
                f"Expected destroy_models to be a bool, received: {type(destroy_models)}"
            )

        if destroy_storage is not None and not isinstance(destroy_storage, bool):
            raise Exception(
                f"Expected destroy_storage to be a bool, received: {type(destroy_storage)}"
            )

        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if max_wait is not None and not isinstance(max_wait, int):
            raise Exception(
                f"Expected max_wait to be a int, received: {type(max_wait)}"
            )

        if model_timeout is not None and not isinstance(model_timeout, int):
            raise Exception(
                f"Expected model_timeout to be a int, received: {type(model_timeout)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="DestroyController", version=12, params=_params
        )
        _params["destroy-models"] = destroy_models
        _params["destroy-storage"] = destroy_storage
        _params["force"] = force
        _params["max-wait"] = max_wait
        _params["model-timeout"] = model_timeout
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CloudSpecResult)
    async def GetCloudSpec(self):
        """GetCloudSpec constructs the CloudSpec for a validated and authorized model.

        Returns -> CloudSpecResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="GetCloudSpec", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(UserAccessResults)
    async def GetControllerAccess(self, entities=None):
        """GetControllerAccess returns the level of access the specified users
        have on the controller.

        entities : typing.Sequence[~Entity]
        Returns -> UserAccessResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="GetControllerAccess", version=12, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(HostedModelConfigsResults)
    async def HostedModelConfigs(self):
        """HostedModelConfigs returns all the information that the client needs in
        order to connect directly with the host model's provider and destroy it
        directly.

        Returns -> HostedModelConfigsResults
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="HostedModelConfigs", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringResult)
    async def IdentityProviderURL(self):
        """IdentityProviderURL returns the URL of the configured external identity
        provider for this controller or an empty string if no external identity
        provider has been configured when the controller was bootstrapped.

        NOTE: the implementation intentionally does not check for SuperuserAccess
        as the URL is known even to users with login access.

        Returns -> StringResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="IdentityProviderURL", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(InitiateMigrationResults)
    async def InitiateMigration(self, specs=None):
        """InitiateMigration attempts to begin the migration of one or
        more models to other controllers.

        specs : typing.Sequence[~MigrationSpec]
        Returns -> InitiateMigrationResults
        """
        if specs is not None and not isinstance(specs, (bytes, str, list)):
            raise Exception(f"Expected specs to be a Sequence, received: {type(specs)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="InitiateMigration", version=12, params=_params
        )
        _params["specs"] = specs
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelBlockInfoList)
    async def ListBlockedModels(self):
        """ListBlockedModels returns a list of all models on the controller
        which have a block in place.  The resulting slice is sorted by model
        name, then owner. Callers must be controller administrators to retrieve the
        list.

        Returns -> ModelBlockInfoList
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="ListBlockedModels", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ModelStatusResults)
    async def ModelStatus(self, entities=None):
        """ModelStatus returns a summary of the model.

        entities : typing.Sequence[~Entity]
        Returns -> ModelStatusResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Controller", request="ModelStatus", version=12, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def ModifyControllerAccess(self, changes=None):
        """ModifyControllerAccess changes the model access granted to users.

        changes : typing.Sequence[~ModifyControllerAccess]
        Returns -> ErrorResults
        """
        if changes is not None and not isinstance(changes, (bytes, str, list)):
            raise Exception(
                f"Expected changes to be a Sequence, received: {type(changes)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller",
            request="ModifyControllerAccess",
            version=12,
            params=_params,
        )
        _params["changes"] = changes
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringResult)
    async def MongoVersion(self):
        """MongoVersion allows the introspection of the mongo version per controller

        Returns -> StringResult
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="MongoVersion", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def RemoveBlocks(self, all_=None):
        """RemoveBlocks removes all the blocks in the controller.

        all_ : bool
        Returns -> None
        """
        if all_ is not None and not isinstance(all_, bool):
            raise Exception(f"Expected all_ to be a bool, received: {type(all_)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="RemoveBlocks", version=12, params=_params
        )
        _params["all"] = all_
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(SummaryWatcherID)
    async def WatchAllModelSummaries(self):
        """WatchAllModelSummaries starts watching the summary updates from the cache.
        This method is superuser access only, and watches all models in the
        controller.

        Returns -> SummaryWatcherID
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller",
            request="WatchAllModelSummaries",
            version=12,
            params=_params,
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(AllWatcherId)
    async def WatchAllModels(self):
        """WatchAllModels starts watching events for all models in the
        controller. The returned AllWatcherId should be used with Next on the
        AllModelWatcher endpoint to receive deltas.

        Returns -> AllWatcherId
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="WatchAllModels", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(NotifyWatchResults)
    async def WatchCloudSpecsChanges(self, entities=None):
        """WatchCloudSpecsChanges returns a watcher for cloud spec changes.

        entities : typing.Sequence[~Entity]
        Returns -> NotifyWatchResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller",
            request="WatchCloudSpecsChanges",
            version=12,
            params=_params,
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(SummaryWatcherID)
    async def WatchModelSummaries(self):
        """WatchModelSummaries starts watching the summary updates from the cache.
        Only models that the user has access to are returned.

        Returns -> SummaryWatcherID
        """
        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Controller", request="WatchModelSummaries", version=12, params=_params
        )

        reply = await self.rpc(msg)
        return reply



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example demonstrate how status works"""

import asyncio
import logging
import sys
from logging import getLogger

from juju.model import Model
from juju.status import formatted_status

LOG = getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


async def main():
    model = Model()
    await model.connect_current()

    application = await model.deploy(
        "ch:ubuntu",
        application_name="ubuntu",
        series="jammy",
        channel="stable",
    )
    await asyncio.sleep(10)
    # Print the status to observe the evolution
    # during a minute
    for _ in range(12):
        try:
            # By setting raw to True, the returned
            # entry contains a FullStatus object with
            # all the available status data.
            # status = await model.status(raw=True)
            status = await formatted_status(model)
            print(status)
        except Exception as e:
            print(e)
        await asyncio.sleep(5)

    print("Removing ubuntu")
    await application.remove()

    print("Disconnecting from model")
    await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Creates a model on the current controller
2. Deploys a charm to it.
3. Attempts to ssh into the charm

"""

import asyncio
import uuid
from logging import getLogger

from juju import utils
from juju.controller import Controller

LOG = getLogger(__name__)


async def main():
    controller = Controller()
    print("Connecting to controller")
    # connect to current controller with current user, per Juju CLI
    await controller.connect()

    try:
        model_name = f"addmodeltest-{uuid.uuid4()}"
        print(f"Adding model {model_name}")
        model = await controller.add_model(model_name)

        print("Deploying ubuntu")
        application = await model.deploy(
            "ch:ubuntu",
            application_name="ubuntu",
            series="jammy",
            channel="stable",
        )

        print("Waiting for active")
        await asyncio.sleep(10)
        await model.wait_for_idle(status="active")

        print("Verifying that we can ssh into the created model")
        ret = await utils.execute_process(
            "juju", "ssh", "-m", model_name, "ubuntu/0", "ls /", log=LOG
        )
        assert ret

        print("Removing ubuntu")
        await application.remove()

        print("Destroying model")
        await controller.destroy_model(model.info.uuid)

    except Exception:
        LOG.exception(f"Test failed! Model {model_name} may not be cleaned up")

    finally:
        print("Disconnecting from controller")
        if model:
            await model.disconnect()
        await controller.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import os

PYLIBJUJU_DEV_FEATURE_FLAG = "PYLIBJUJU_DEV_FEATURE_FLAGS"


DEFAULT_VALUES_FLAG = "default_values"


def feature_enabled(name):
    flags = os.environ.get(PYLIBJUJU_DEV_FEATURE_FLAG)
    if flags is not None:
        parts = [s.strip() for s in flags.split(",")]
        return name in parts
    return False



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example shows how to deploy a local charm. It:

1. Connects to current model.
2. Uploads a local charm (directory on filesystem) to the model.
3. Deploys the uploaded charm.

"""

import asyncio
import logging

from juju.model import Model


async def main():
    model = Model()
    await model.connect()

    # Deploy a local charm using a path to the charm directory
    await model.deploy(
        "./charms/ubuntu",
        application_name="ubuntu",
        series="trusty",
    )

    await model.disconnect()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    ws_logger = logging.getLogger("websockets.protocol")
    ws_logger.setLevel(logging.INFO)
    asyncio.run(main())
